import datetime
import numpy as np
from scipy.optimize import minimize, OptimizeResult
from scipy.stats import f
from pyERGM.sampling import NaiveMetropolisHastings

from pyERGM.metrics import *


class ERGM():
    def __init__(self,
                 n_nodes,
                 metrics_collection: Collection[Metric],
                 is_directed: bool,
                 initial_thetas: dict = None,
                 initial_normalization_factor=None,
                 seed_MCMC_proba=0.25,
                 verbose=True,
                 use_sparse_matrix=False,
                 fix_collinearity=True,
                 collinearity_fixer_sample_size=1000,
                 is_distributed_optimization=False,
                 optimization_options={},
                 **kwargs):
        """
        An ERGM model object. 
        
        Parameters
        ----------
        n_nodes : int 
            Number of nodes in the network.

        metrics_collection : Collection[Metric] 
            A list of Metric objects for calculating statistics of a network.

        is_directed : bool 
            Whether the graph is directed or not.

        initial_thetas : npdarray 
            Optional. The initial values of the coefficients of the ERGM. If not provided, they are randomly initialized.

        initial_normalization_factor : float 
            Optional. The initial value of the normalization factor. If not provided, it is randomly initialized.

        seed_MCMC_proba : float 
            Optional. The probability of a connection in the seed network for MCMC sampling, in case no seed network is provided. *Defaults to 0.25*

        sample_size : int 
            Optional. The number of networks to sample via MCMC. If number of samples is odd, it is increased by 1. This is because downstream algorithms assume the sample size is even (e.g. the Covariance matrix estimation). *Defaults to 1000*

        use_sparse_matrix : bool
            Optional. Whether to use sparse matrices for the adjacency matrix. 
            Sparse matrices are implemented via PyTorch's Sparse Tensor's, which are still in beta.  *Defaults to False*
        
        fix_collinearity : bool
            Optional. Whether to fix collinearity in the metrics. *Defaults to True*

        collinearity_fixer_sample_size : int
            Optional. The number of networks to sample for fixing collinearity. *Defaults to 1000*
        
        """
        self._n_nodes = n_nodes
        self._is_directed = is_directed
        self._is_distributed_optimization = is_distributed_optimization

        self._metrics_collection = MetricsCollection(metrics_collection, self._is_directed, self._n_nodes,
                                                     use_sparse_matrix=use_sparse_matrix,
                                                     fix_collinearity=fix_collinearity and (initial_thetas is None),
                                                     collinearity_fixer_sample_size=collinearity_fixer_sample_size,
                                                     is_collinearity_distributed=self._is_distributed_optimization,
                                                     num_samples_per_job_collinearity_fixer=kwargs.get(
                                                         'num_samples_per_job_collinearity_fixer', 5))

        self.n_node_features = self._metrics_collection.n_node_features
        self.node_feature_names = self._metrics_collection.node_feature_names.copy()
        self.node_features_n_categories = self._metrics_collection.node_features_n_categories.copy()

        if initial_thetas is not None:
            if type(initial_thetas) != dict:
                raise ValueError("Initial thetas must be a dictionary keyed by feature names, as returned by "
                                 "`ERGM.get_model_parameters`")
            self._thetas = np.zeros(self._metrics_collection.calc_num_of_features())
            current_model_params = self.get_model_parameters()
            if len(set(initial_thetas.keys()).difference(set(current_model_params.keys()))) > 0:
                raise ValueError("Got initial thetas that do not match the collection of Metrics!")

            total_num_features = len(current_model_params.keys())
            # Iterating the reversed list of keys for not interfering with indexing: always remove the last feature to
            # be removed, thus not changing the indices of features needed to be removed with smaller indices.
            for rev_feat_idx, feat_name in enumerate(list(current_model_params.keys())[::-1]):
                if feat_name not in initial_thetas.keys():
                    self._metrics_collection.remove_feature_by_idx(total_num_features - rev_feat_idx - 1)

            current_model_params = self.get_model_parameters()
            for feat_name in current_model_params.keys():
                current_model_params[feat_name] = initial_thetas[feat_name]
            self._thetas = np.array(list(current_model_params.values()))
        else:
            self._thetas = self._get_random_thetas(sampling_method="uniform")

        if initial_normalization_factor is not None:
            self._normalization_factor = initial_normalization_factor
        else:
            self._normalization_factor = np.random.normal(50, 10)

        self._seed_MCMC_proba = seed_MCMC_proba

        self.optimization_start_time = None

        self.verbose = verbose
        self.optimization_options = optimization_options

        self._exact_average_mat = None

        self.mh_sampler = NaiveMetropolisHastings(self._thetas, self._metrics_collection)

    def print_model_parameters(self):
        """
        Prints the parameters of the ERGM model.
        """
        print(f"Number of nodes: {self._n_nodes}")
        print(f"Thetas: {self._thetas}")
        print(f"Normalization factor approx: {self._normalization_factor}")
        print(f"Is directed: {self._is_directed}")

    def calculate_weight(self, W: np.ndarray):
        if len(W.shape) != 2 or W.shape[0] != self._n_nodes or W.shape[1] < self._n_nodes:
            raise ValueError(f"The dimensions of the given adjacency matrix, {W.shape}, don't comply with the number of"
                             f" nodes in the network: {self._n_nodes}")
        features = self._metrics_collection.calculate_statistics(W)
        weight = np.exp(np.dot(self._thetas, features))

        return weight

    def _get_random_thetas(self, sampling_method="uniform"):
        if sampling_method == "uniform":
            return np.random.uniform(-1, 1, self._metrics_collection.num_of_features)
        else:
            raise ValueError(f"Sampling method {sampling_method} not supported. See docs for supported samplers.")

    def generate_networks_for_sample(self,
                                     sample_size,
                                     seed_network=None,
                                     replace=True,
                                     burn_in=10000,
                                     mcmc_steps_per_sample=1000,
                                     sampling_method="metropolis_hastings",
                                     edge_proposal_method='uniform',
                                     edge_node_flip_ratio=None
                                     ):
        if sampling_method == "metropolis_hastings":
            if seed_network is None:
                G = nx.erdos_renyi_graph(self._n_nodes, self._seed_MCMC_proba, directed=self._is_directed)
                seed_connectivity_matrix = nx.to_numpy_array(G)
                seed_neuron_features = np.zeros((self._n_nodes, self.n_node_features))
                for feature_name, feature_indices in self.node_feature_names.items():
                    for feature_index in feature_indices:
                        seed_neuron_features[:, feature_index] = np.random.choice(self.node_features_n_categories[feature_name], size=self._n_nodes)
                seed_network = np.concatenate((seed_connectivity_matrix, seed_neuron_features), axis=1)

            self.mh_sampler.set_thetas(self._thetas)
            return self.mh_sampler.sample(seed_network, sample_size, replace=replace, burn_in=burn_in,
                                          steps_per_sample=mcmc_steps_per_sample,
                                          edge_proposal_method=edge_proposal_method,
                                          node_feature_names=self.node_feature_names,
                                          node_features_n_categories=self.node_features_n_categories,
                                          edge_node_flip_ratio=edge_node_flip_ratio)
        elif sampling_method == "exact":
            return self._generate_exact_sample(sample_size)
        else:
            raise ValueError(f"Unrecognized sampling method {sampling_method}")

    @staticmethod
    def do_estimate_covariance_matrix(optimization_method, convergence_criterion):
        if optimization_method == "newton_raphson" or convergence_criterion == "hotelling":
            return True
        return False

    def _mple_fit(self, observed_network, optimization_method: str = 'L-BFGS-B', **kwargs):
        """
        Perform MPLE estimation of the ERGM parameters.
        This is done by fitting a logistic regression model, where the X values are the change statistics
        calculated for every edge in the observed network, and the predicted values are the presence of the edge.

        More precisely, we create a train dataset with (X_{i, j}, y_{i,j}) where -
            X_{i, j} = g(y_{i,j}^+) - g(y_{i, j}^-)
            y_{i, j} = 1 if there is an edge between i and j, 0 otherwise

        Parameters
        ----------
        observed_network : np.ndarray
            The adjacency matrix of the observed network.
        
        Returns
        -------
        thetas: np.ndarray
            The estimated coefficients of the ERGM.
        """
        print("MPLE")
        trained_thetas, prediction, success = mple_logistic_regression_optimization(self._metrics_collection,
                                                                           observed_network,
                                                                           is_distributed=self._is_distributed_optimization,
                                                                           optimization_method=optimization_method,
                                                                           num_edges_per_job=kwargs.get(
                                                                               "num_edges_per_job", 100000))

        self._exact_average_mat = self._rearrange_prediction_to_av_mat(prediction)

        return trained_thetas, success

    def _rearrange_prediction_to_av_mat(self, prediction):
        av_mat = np.zeros((self._n_nodes, self._n_nodes))
        if self._is_directed:
            av_mat[~np.eye(self._n_nodes, dtype=bool)] = prediction
        else:
            upper_triangle_indices = np.triu_indices(self._n_nodes, k=1)
            av_mat[upper_triangle_indices] = prediction
            lower_triangle_indices_aligned = (upper_triangle_indices[1], upper_triangle_indices[0])
            av_mat[lower_triangle_indices_aligned] = prediction
        return av_mat

    # TODO: decide how a getter for self._exact_average_mat fits in now - if the model is dyadic
    #  independent, the observed_network doesn't matter - predictions will be always the same, and this is the exact
    #  average matrix of the model, so should be computed once and stored. If the model is dyadic dependent, this is an
    #  approximation, and the degree to which changes in the observed_network will change the prediction depend on the
    #  metrics and the specific networks, it can not be pre-determined.
    def get_mple_prediction(self, observed_network: np.ndarray):
        is_dyadic_independent = not self._metrics_collection._has_dyadic_dependent_metrics
        if is_dyadic_independent and self._exact_average_mat is not None:
            return self._exact_average_mat.copy()

        # TODO: handle this case
        if self._is_distributed_optimization:
            raise NotImplementedError(
                "calculating mple predictions distributed without fitting is yet to be implemented")
        Xs, _ = self._metrics_collection.prepare_mple_data(observed_network)
        pred = calc_logistic_regression_predictions(Xs, self._thetas).flatten()
        if is_dyadic_independent:
            self._exact_average_mat = self._rearrange_prediction_to_av_mat(pred)
            return self._exact_average_mat.copy()
        return self._rearrange_prediction_to_av_mat(pred)

    def _do_MPLE(self, theta_init_method):
        if not self._metrics_collection._has_dyadic_dependent_metrics or theta_init_method == "mple":
            return True
        return False

    def _generate_exact_sample(self, sample_size: int = 1):
        # TODO: support getting a flag of `replace` which will enable sampling with no replacements (generating samples
        #  of different networks).
        if self._metrics_collection._has_dyadic_dependent_metrics:
            raise ValueError("Cannot sample exactly from a model that is dyadic dependent!")
        if self._exact_average_mat is None:
            raise ValueError("Cannot sample exactly from a model that is not trained! Call `model.fit()` and pass an "
                             "observed network!")

        return sample_from_independent_probabilities_matrix(self._exact_average_mat, sample_size)

    def fit(self, observed_network,
            lr=0.1,
            opt_steps=1000,
            steps_for_decay=100,
            lr_decay_pct=0.01,
            l2_grad_thresh=0.001,
            sliding_grad_window_k=10,
            max_sliding_window_size=100,
            max_nets_for_sample=1000,
            sample_pct_growth=0.02,
            optimization_method="newton_raphson",
            convergence_criterion="model_bootstrap",
            cov_matrix_estimation_method="naive",
            cov_matrix_num_batches=25,
            hotelling_confidence=0.99,
            theta_init_method="mple",
            no_mple=False,
            mcmc_burn_in=1000,
            mcmc_seed_network=None,
            mcmc_steps_per_sample=10,
            mcmc_sample_size=100,
            edge_proposal_method='uniform',
            edge_node_flip_ratio=None,
            observed_node_features=None,
            **kwargs
            ):
        """
        Fit an ERGM model to a given network with one of the two fitting methods - MPLE or MCMLE.

        Parameters
        ----------
        observed_network : np.ndarray
            The observed network connectivity matrix, with shape (n, n).
        
        observed_node_features: dict
            Optional. A dictionary of node features. Each key is the name of the feature, and the value is a list of
            `n` numbers representing the feature of every node. *Defaults to None*.
            
            e.g. - `observed_node_features = {"E_I": [[0, 1, 1, 1]]}`
            
        lr : float
            Optional. The learning rate for the optimization. *Defaults to 0.1*

        opt_steps : int
            Optional. The number of optimization steps to run. *Defaults to 1000*

        steps_for_decay : int
            Optional. The number of steps after which to decay the optimization params. 
            *Defaults to 100* # TODO - redundant parameter?

        lr_decay_pct : float
            Optional. The decay factor for the learning rate. *Defaults to 0.01*

        l2_grad_thresh : float
            Optional. The threshold for the L2 norm of the gradient to stop the optimization. 
            Relevant only for convergence criterion "zero_grad_norm". *Defaults to 0.001*

        sliding_grad_window_k : int
            Optional. The size of the sliding window for the gradient, for which we use to calculate the mean gradient norm. 
            This value is then tested against l2_grad_thresh to decide whether optimization halts.
            Relevant only for convergence criterion "zero_grad_norm". *Defaults to 10*

        max_sliding_window_size : int
            Optional. The maximum size of the sliding window for the gradient. 
            Relevant only for convergence criterion "zero_grad_norm". *Defaults to 100*

        max_nets_for_sample : int
            Optional. The maximum number of networks to sample with MCMC. *Defaults to 1000*
            #TODO - Do we still need this? Seems like increasing the sample size isn't necessary (we'll gonna pick large sample sizes anyway)
            
        sample_pct_growth : float
            Optional. The percentage growth of the number of networks to sample, which we want to increase over time.
            *Defaults to 0.02*. #TODO - Same as `max_nets_for_sample`. Do we still need this?

        optimization_method : str
            Optional. The optimization method to use. Can be either "newton_raphson" or "gradient_descent".
            *Defaults to "newton_raphson"*.

        convergence_criterion : str
            Optional. The criterion for convergence. Can be either "hotelling" or "zero_grad_norm".
            *Defaults to "zero_grad_norm"*.
            # TODO - Revisit this when we fix convergence criterion.
            
        cov_matrix_estimation_method : str
            Optional. The method to estimate the covariance matrix. 
            Supported methods - `naive`, `batch`, `multivariate_initial_sequence`. *Defaults to "batch"*.
        
        cov_matrix_num_batches : int
            Optional. The number of batches to use for estimating the covariance matrix.
            Relevant only for `cov_matrix_estimation_method="batch"`. *Defaults to 25*.

        hotelling_confidence : float
            Optional. The confidence level for the Hotelling's T-squared test. *Defaults to 0.99*.
        
        theta_init_method : str
            Optional. The method to initialize the theta values. Can be either "uniform", "mple" or "use_existing" (which uses the current thetas).
            The MPLE method can be used even for dyadic dependent models, since it serves as a good starting point for the MCMLE.
            *Defaults to "mple"*.

        no_mple : bool
            Optional. Whether to skip the MPLE step and go directly to MCMLE. *Defaults to False*.
        
        mcmc_burn_in : int
            Optional. The number of burn-in steps for the MCMC sampler. *Defaults to 1000*.
        
        mcmc_steps_per_sample : int
            Optional. The number of steps to run the MCMC sampler for each sample. *Defaults to 10*.
        
        mcmc_seed_network : np.ndarray
            Optional. The seed network for the MCMC sampler. If not provided, the thetas that are currently set are used to 
            calculate the MPLE prediction, from which the sample is drawn. *Defaults to None*.
        
        mcmc_sample_size : int
            Optional. The number of networks to sample with the MCMC sampler. *Defaults to 100*.

        edge_proposal_method : str
            Optional. The method for the MCMC proposal distribution. This is defined as a distribution over the edges
            of the network, which implies how to choose a proposed graph out of all graphs that are 1-edge-away from the
            current graph. *Defaults to "uniform"*.
        
        Returns 
        -------
        (grads, hotelling_statistics) : (np.ndarray, list)
        # TODO - what do we want to return?
        """
        # Create the full observed network from adjacency matrix and node features:
        if observed_node_features is not None:
            ordered_observed_node_features = [observed_node_features[fname] for fname in self.node_feature_names.keys()]
            ordered_observed_node_features = [one_d_f for f in ordered_observed_node_features for one_d_f in f]
            ordered_observed_node_features = np.array(ordered_observed_node_features).T
            observed_network = np.concatenate([observed_network, ordered_observed_node_features], axis=1)

        # This is because we assume the sample size is even when estimating the covariance matrix (in
        # calc_capital_gammas).
        if mcmc_sample_size % 2 != 0:
            mcmc_sample_size += 1

        if theta_init_method == "use_existing":
            print(f"Using existing thetas")
            pass
        elif not no_mple and self._do_MPLE(theta_init_method):
            self._thetas, success = self._mple_fit(observed_network,
                                          optimization_method=kwargs.get('mple_optimization_method', 'L-BFGS-B'),
                                          num_edges_per_job=kwargs.get('num_edges_per_job', 100000))

            if not self._metrics_collection._has_dyadic_dependent_metrics:
                print(f"Model is dyadic independent - using only MPLE instead of MCMLE")
                return {"success": success}

        elif theta_init_method == "uniform":
            self._thetas = self._get_random_thetas(sampling_method="uniform")
        else:
            raise ValueError(f"Theta initialization method {theta_init_method} not supported")

        if optimization_method not in ["newton_raphson", "gradient_descent"]:
            raise ValueError(f"Optimization method {optimization_method} not supported.")

        # As in the constructor, the sample size must be even.
        if max_nets_for_sample % 2 != 0:
            max_nets_for_sample += 1

        if convergence_criterion == "observed_bootstrap":
            for metric in self._metrics_collection.metrics:
                if not hasattr(metric, "calculate_bootstrapped_features"):
                    raise ValueError(
                        f"metric {metric.name} does not have a calculate_bootstrapped_features method, the "
                        f"model doesn't support observed_bootstrap as a convergence criterion.")
            num_subsamples_data = kwargs.get("num_subsamples_data", 1000)
            data_splitting_method = kwargs.get("data_splitting_method", "uniform")
            bootstrapped_features = self._metrics_collection.bootstrap_observed_features(observed_network,
                                                                                         num_subsamples=num_subsamples_data,
                                                                                         splitting_method=data_splitting_method)
            observed_covariance = covariance_matrix_estimation(bootstrapped_features,
                                                               bootstrapped_features.mean(axis=1), method='naive')
            inv_observed_covariance = np.linalg.inv(observed_covariance)

        print(f"Initial thetas - {self._thetas}")
        print(f"{datetime.datetime.now()} optimization started")

        self.optimization_start_time = time.time()
        num_of_features = self._metrics_collection.num_of_features

        grads = np.zeros((opt_steps, num_of_features))

        if mcmc_seed_network is None and not no_mple:
            probabilities_matrix = self.get_mple_prediction(observed_network)
            mcmc_seed_network = sample_from_independent_probabilities_matrix(probabilities_matrix, 1)
            mcmc_seed_network = mcmc_seed_network[:, :, 0]
        burn_in = mcmc_burn_in
        for i in range(opt_steps):
            if i > 0:
                burn_in = 0
            networks_for_sample = self.generate_networks_for_sample(sample_size=mcmc_sample_size,
                                                                    seed_network=mcmc_seed_network, burn_in=burn_in,
                                                                    mcmc_steps_per_sample=mcmc_steps_per_sample,
                                                                    edge_proposal_method=edge_proposal_method,
                                                                    edge_node_flip_ratio=edge_node_flip_ratio)
            mcmc_seed_network = networks_for_sample[:, :, -1]

            features_of_net_samples = self._metrics_collection.calculate_sample_statistics(networks_for_sample)
            mean_features = np.mean(features_of_net_samples, axis=1)
            observed_features = self._metrics_collection.calculate_statistics(observed_network)

            grad = calc_nll_gradient(observed_features, mean_features)
            if ERGM.do_estimate_covariance_matrix(optimization_method, convergence_criterion):
                # This is for allowing numba to compile and pickle the large function
                sys.setrecursionlimit(2000)
                print(f"{datetime.datetime.now()} Started estimating cov matrix ")
                estimated_cov_matrix = covariance_matrix_estimation(features_of_net_samples,
                                                                    mean_features,
                                                                    method=cov_matrix_estimation_method,
                                                                    num_batches=cov_matrix_num_batches)
                print(f"{datetime.datetime.now()} Done estimating cov matrix ")
                inv_estimated_cov_matrix = np.linalg.pinv(estimated_cov_matrix)
                print(f"{datetime.datetime.now()} Done inverting cov matrix ")
            if optimization_method == "newton_raphson":
                self._thetas = self._thetas - lr * inv_estimated_cov_matrix @ grad

            elif optimization_method == "gradient_descent":
                self._thetas = self._thetas - lr * grad

            grads[i] = grad

            idx_for_sliding_grad = np.max([0, i - sliding_grad_window_k + 1])
            sliding_window_grads = grads[idx_for_sliding_grad:i + 1].mean()

            if (i + 1) % steps_for_decay == 0:
                delta_t = time.time() - self.optimization_start_time
                print(
                    f"Step {i + 1} - lr: {lr:.7f}, time from start: {delta_t:.2f}, window_grad: {sliding_window_grads:.2f}")

                lr *= (1 - lr_decay_pct)

                print(self._thetas)

                if mcmc_sample_size < max_nets_for_sample:
                    mcmc_sample_size *= (1 + sample_pct_growth)
                    mcmc_sample_size = np.min([int(mcmc_sample_size), max_nets_for_sample])
                    # As in the constructor, the sample size must be even.
                    if mcmc_sample_size % 2 != 0:
                        mcmc_sample_size += 1
                    print(f"\t Sample size increased at step {i + 1} to {mcmc_sample_size}")

                if sliding_grad_window_k < max_sliding_window_size:
                    sliding_grad_window_k *= (1 + sample_pct_growth)
                    sliding_grad_window_k = np.min(
                        [np.ceil(sliding_grad_window_k).astype(int), max_sliding_window_size])

            print(f"{datetime.datetime.now()} Starting to test for convergence")
            convergence_tester = ConvergenceTester()

            if convergence_criterion == "hotelling":
                print(f"{datetime.datetime.now()} \t Starting a `hotelling` test")
                convergence_results = convergence_tester.hotelling(observed_features, mean_features, inv_estimated_cov_matrix, mcmc_sample_size, hotelling_confidence)
                print(f"{datetime.datetime.now()} \t DONE with `hotelling` test")
                if convergence_results["success"]:
                    print(f"Reached a confidence of {hotelling_confidence} with the hotelling convergence test! DONE! ")
                    grads = grads[:i]
                    break

            elif convergence_criterion == "zero_grad_norm":
                if np.linalg.norm(sliding_window_grads) <= l2_grad_thresh:
                    print(f"Reached threshold of {l2_grad_thresh} after {i} steps. DONE!")
                    grads = grads[:i]

                    # TODO - implement `convergence_results` for this kind of convergence
                    break
                
            elif convergence_criterion == "observed_bootstrap":
                confidence = kwargs.get("bootstrap_convergence_confidence", 0.95)
                stds_away_thr = kwargs.get("bootstrap_convergence_stds_away_thr", 1)

                convergence_results = convergence_tester.bootstrapped_mahalanobis_from_observed(
                    observed_features,
                    features_of_net_samples,
                    inv_observed_covariance,
                    num_subsamples=kwargs.get("num_model_sub_samples", 100),
                    subsample_size=kwargs.get("model_subsample_size", 1000),
                    confidence=confidence,
                    stds_away_thr=stds_away_thr,
                )

                if convergence_results["success"]:
                    print(f"Reached a confidence of {confidence} with the bootstrap convergence "
                          f"test! The model is likely to be up to {stds_away_thr} stds from "
                          f"the data, according to the estimated data variability DONE! ")
                    grads = grads[:i]
                    break
            elif convergence_criterion == "model_bootstrap": 
                print(f"{datetime.datetime.now()} \t Starting a `model_bootstrap` test")           
                convergence_results = convergence_tester.bootstrapped_mahalanobis_from_model(
                    observed_features,
                    features_of_net_samples,
                    num_subsamples=kwargs.get("num_model_sub_samples", 100),
                    subsample_size=kwargs.get("model_subsample_size", 1000),
                    confidence=kwargs.get("bootstrap_convergence_confidence", 0.95),
                    stds_away_thr=kwargs.get("bootstrap_convergence_stds_away_thr", 1),
                )     
                print(f"{datetime.datetime.now()} \t DONE with `model_bootstrap` test")
                if convergence_results["success"]:
                    print(f"Reached a confidence of {kwargs.get('bootstrap_convergence_confidence', 0.95)} with the bootstrap convergence "
                          f"test! The model is likely to be up to {kwargs.get('bootstrap_convergence_stds_away_thr', 1)} stds from "
                          f"the data, according to the estimated data variability DONE! ")
                    grads = grads[:i]
                    break
            else:
                raise ValueError(f"Convergence criterion {convergence_criterion} not defined")
        
            sys.stdout.flush()

        self._last_mcmc_chain_features = features_of_net_samples

        return convergence_results

    def get_mcmc_diagnostics(self, sampled_networks=None, observed_network=None):
        """
        Get MCMC diagnostics for the last chain or a sample of networks.

        Parameters
        ----------
        sampled_networks : np.ndarray
            Optional. A (n, n+k, sample_size) tensor of sampled networks. If not provided, the last chain is used.
        observed_network : np.ndarray
            Optional. The network matrix of the observed network. If provided, the traces of the features across the
            chain are normalized based on the observed network features.
        """
        if sampled_networks is not None:
            print(f"Calculating MCMC diagnostics for a sample of {sampled_networks.shape[2]} networks")
            features = self._metrics_collection.calculate_sample_statistics(sampled_networks)
        elif self._last_mcmc_chain_features is not None:
            print(
                f"Calculating MCMC diagnostics for the last chain, with {self._last_mcmc_chain_features.shape[1]} networks")
            features = self._last_mcmc_chain_features.copy()
        else:
            # TODO: decide what to do in this case. Maybe give the user an option to generate a new chain.
            raise ValueError(
                "No sampled networks provided and no last chain found. Either rerun with a `sampled_networks` parameter, or run the `fit` function first.")

        features_mean = np.mean(features, axis=1)
        features_std = np.std(features, axis=1)

        # Normalize the features based on the observed network
        if observed_network is not None:
            observed_features = self._metrics_collection.calculate_statistics(observed_network)
            features = features - observed_features[:, None]

        trace_per_features = {}
        parameter_names = self._metrics_collection.get_parameter_names()
        for i, feature in enumerate(parameter_names):
            trace = features[i, :]
            trace_per_features[str(feature)] = trace

        quantlies_to_calculate = [0.025, 0.25, 0.5, 0.75, 0.975]
        quantiles = np.quantile(features, quantlies_to_calculate, axis=1)

        quantile_values = {k: q for k, q in zip(quantlies_to_calculate, quantiles)}

        return {
            "trace": trace_per_features,
            "quantiles": quantile_values,
            "mean": features_mean,
            "std": features_std
        }

    def calculate_probability(self, W: np.ndarray):
        """
        Calculate the probability of a graph under the ERGM model.

        Parameters
        ----------
        W : np.ndarray
            A connectivity matrix.

        Returns
        -------
        prob : float
            The probability of the graph under the ERGM model.
        """

        if self._normalization_factor is None or self._thetas is None:
            raise ValueError("Normalization factor and thetas not set, fit the model before calculating probability.")

        weight = self.calculate_weight(W)
        prob = weight / self._normalization_factor

        return prob

    def get_model_parameters(self):
        parameter_names = self._metrics_collection.get_parameter_names()
        return dict(zip(parameter_names, self._thetas))

    def get_ignored_features(self):
        return self._metrics_collection.get_ignored_features()

    def get_mple_prediction(self, input_graph):
        Xs, ys = self._metrics_collection.prepare_mple_data(input_graph)

        prediction = calc_logistic_regression_predictions(Xs, self._thetas)
        exact_average_mat = np.zeros((self._n_nodes, self._n_nodes))

        if self._is_directed:
            exact_average_mat[~np.eye(self._n_nodes, dtype=bool)] = prediction

        return exact_average_mat


class BruteForceERGM(ERGM):
    """
    A class that implements ERGM by iterating over the entire space of networks and calculating stuff exactly (rather
    than using statistical methods for approximating and sampling).
    This is mainly for tests.
    """
    # The maximum number of nodes that is allowed for carrying brute force calculations (i.e. iterating the whole space
    # of networks and calculating stuff exactly). This becomes not tractable above this limit.
    MAX_NODES_BRUTE_FORCE_DIRECTED = 5
    MAX_NODES_BRUTE_FORCE_NOT_DIRECTED = 7

    def __init__(self,
                 n_nodes,
                 metrics_collection: Collection[Metric],
                 is_directed=False,
                 initial_thetas=None):
        super().__init__(n_nodes,
                         metrics_collection,
                         is_directed,
                         initial_thetas)

        if is_directed:
            num_connections = (n_nodes ** 2 - n_nodes)
        else:
            num_connections = (n_nodes ** 2 - n_nodes) // 2

        self._num_nets = 2 ** num_connections

        self._num_features = self._metrics_collection.num_of_features

        all_networks = np.zeros((self._n_nodes, self._n_nodes, self._num_nets))
        for i in range(self._num_nets):
            all_networks[:, :, i] = construct_adj_mat_from_int(i, self._n_nodes, is_directed=is_directed)

        self.all_features_by_all_nets = self._metrics_collection.calculate_sample_statistics(all_networks)

        self._all_weights = np.exp(np.sum(self._thetas[:, None] * self.all_features_by_all_nets, axis=0))
        self._normalization_factor = self._all_weights.sum()

    def _validate_net_size(self):
        return (
                (self._is_directed and self._n_nodes <= BruteForceERGM.MAX_NODES_BRUTE_FORCE_DIRECTED)
                or
                (not self._is_directed and self._n_nodes <= BruteForceERGM.MAX_NODES_BRUTE_FORCE_NOT_DIRECTED)
        )

    def calculate_weight(self, W: np.ndarray):
        adj_mat_idx = construct_int_from_adj_mat(W, self._is_directed)
        return self._all_weights[adj_mat_idx]

    def sample_network(self, sampling_method="Exact", seed_network=None, steps=0):
        if sampling_method != "Exact":
            raise ValueError("BruteForceERGM supports only exact sampling (this is its whole purpose)")

        all_nets_probs = self._all_weights / self._normalization_factor
        sampled_idx = np.random.choice(all_nets_probs.size, p=all_nets_probs)
        return construct_adj_mat_from_int(sampled_idx, self._n_nodes, self._is_directed)

    def fit(self, observed_network):
        def nll(thetas):
            model = BruteForceERGM(self._n_nodes, list(self._metrics_collection.metrics),
                                   initial_thetas={feat_name: thetas[i] for i, feat_name in
                                                   enumerate(self._metrics_collection.get_parameter_names())},
                                   is_directed=self._is_directed)
            return np.log(model._normalization_factor) - np.log(model.calculate_weight(observed_network))

        def nll_grad(thetas):
            model = BruteForceERGM(self._n_nodes, list(self._metrics_collection.metrics),
                                   initial_thetas={feat_name: thetas[i] for i, feat_name in
                                                   enumerate(self._metrics_collection.get_parameter_names())},
                                   is_directed=self._is_directed)
            observed_features = model._metrics_collection.calculate_statistics(observed_network)
            all_probs = model._all_weights / model._normalization_factor

            expected_features = self.all_features_by_all_nets @ all_probs
            return expected_features - observed_features

        def after_iteration_callback(intermediate_result: OptimizeResult):
            self.optimization_iter += 1
            cur_time = time.time()
            print(f'iteration: {self.optimization_iter}, time from start '
                  f'training: {cur_time - self.optimization_start_time} '
                  f'log likelihood: {-intermediate_result.fun}')

        self.optimization_iter = 0
        print("optimization started")
        self.optimization_start_time = time.time()
        res = minimize(nll, self._thetas, jac=nll_grad, callback=after_iteration_callback)
        self._thetas = res.x
        return res


class ConvergenceTester:
    def __init__(self):
        pass
    
    @staticmethod
    def _get_subsample_features(sampled_networks_features, num_subsamples, subsample_size):
        """
        Receives a sample of networks, and subsample for `num_subsamples` times, each time with `subsample_size` networks.
        For each subsample, calculates the sample statistics and reshapes the result to a tensor of shape (num_of_features, num_subsamples, subsample_size).

        Parameters
        ----------
        sampled_networks_features : np.ndarray
            Features of a sample of networks that will be used for subsampling
        
        num_subsamples : int
            The number of subsamples to draw from the sample.
        
        subsample_size : int
            The size of each subsample.
    
        Returns
        -------
        sub_samples_features : np.ndarray
            A tensor of shape (num_of_features, num_subsamples, subsample_size) containing the features of all subsamples.
        """
        sample_size = sampled_networks_features.shape[1]

        sub_sample_indices = np.random.choice(np.arange(sample_size), size=num_subsamples * subsample_size)
        sub_samples_features = sampled_networks_features[:, sub_sample_indices]
        sub_samples_features = sub_samples_features.reshape(-1, num_subsamples, subsample_size)

        return sub_samples_features

    @staticmethod
    def hotelling(observed_features, mean_features, inverted_sample_cov_matrix, sample_size, confidence=0.99):
        """
        Run the Hotelling's T-squared test for convergence.

        The T-Squarted statistic is calculated as - 
            t^2 = n * dist^2
        where dist is the Mahalanobis distance between the observed and the mean features, used the given
        covariance matrix.

        The T^2 statistic can be transformed into an F statistic - 
            F = (n-p / p(n-1)) * t^2
        where p is the number of features and n is the sample size.
        Finally the F statistic is compared to the critical value of the F distribution with p and n-p degrees of freedom.

        Parameters
        ----------
        observed_features : np.ndarray
            The observed features of the network.
        
        mean_features : np.ndarray
            The mean features of the networks sampled from the model.
        
        inverted_sample_cov_matrix : np.ndarray
            The inverted covariance matrix of the features that were calculated from the model sample.
        
        sample_size : int  
            The number of networks sampled from the model.
        
        confidence : float
            The confidence level for the test. *Defaults to 0.99*.
        """
        dist = mahalanobis(observed_features, mean_features, inverted_sample_cov_matrix)
        hotelling_t_statistic = sample_size * dist * dist

        num_of_features = observed_features.shape[0]

        hotelling_as_f_statistic = ((sample_size - num_of_features) / (
                        num_of_features * (sample_size - 1))) * hotelling_t_statistic
    
        hotelling_critical_value = f.ppf(1 - confidence, num_of_features, sample_size - num_of_features)

        return {
            "success": hotelling_as_f_statistic <= hotelling_critical_value,
            "statistic": hotelling_as_f_statistic,
            "threshold": hotelling_critical_value
        }
        
    @staticmethod
    def bootstrapped_mahalanobis_from_observed(
                                    observed_features, 
                                    sampled_networks_features,
                                    inverted_observed_cov_matrix,
                                    num_subsamples=100,
                                    subsample_size=1000,
                                    confidence=0.95,
                                    stds_away_thr=1):
        """
        Repeatedly subsample from a sample of networks, and calculate the distance between the subsample mean
        and the observed features. The distance is calculated using the Mahalanobis distance, with the covariance matrix
        being an estimation of the observed covariance. The observed covariance can either be the real covariance of the data,
        or an estimation of the covariance matrix (using methods like `Network splitting augmentation` or `Noise augmentation`).
        """
        mahalanobis_dists = np.zeros(num_subsamples)

        sub_samples_features = ConvergenceTester._get_subsample_features(sampled_networks_features, num_subsamples, subsample_size)
        mean_per_subsample = sub_samples_features.mean(axis=2)

        for cur_subsam_idx in range(num_subsamples):
            cur_subsample_mean = mean_per_subsample[:, cur_subsam_idx]
            mahalanobis_dists[cur_subsam_idx] = mahalanobis(observed_features, cur_subsample_mean, inverted_observed_cov_matrix)
        
        empirical_threshold = np.quantile(mahalanobis_dists, confidence)

        return {
            "success": empirical_threshold < stds_away_thr,
            "statistic": empirical_threshold,
            "threshold": stds_away_thr
        }

    @staticmethod
    def bootstrapped_mahalanobis_from_model(
                                    observed_features, 
                                    sampled_networks_features,
                                    num_subsamples=100,
                                    subsample_size=1000,
                                    confidence=0.95,
                                    stds_away_thr=1):
        """
        Repeatedly subsample from a collection of networks sampled from the model (`sampled_networks`), and calculate the Mahalanobis distance 
        between each subsample mean and the observed network. This is equivalent to generating multiple estimations of the model mean & covariance.
        We calculate the cutoff threshold for the Mahalanobis distance, according to the provided `confidence` level, and then verify whether
        the empirical threshold is below `stds_away_thr` (which is standard deviations away from the observed data).

        Parameters
        ----------
        observed_features : np.ndarray
            The observed features of the network.
        
        sampled_networks_features : np.ndarray
            Features of networks sampled from the model.
        
        num_subsamples : int
            The number of subsamples to draw. *Defaults to 100*.

        subsample_size : int
            The size of each subsample. *Defaults to 1000*.
        
        confidence : float
            The confidence level for the test. *Defaults to 0.95*.

        stds_away_thr : float
            The desired threshold for the Mahalanobis distance, in units of std *Defaults to 1*.
        """
        mahalanobis_dists = np.zeros(num_subsamples)

        sub_samples_features = ConvergenceTester._get_subsample_features(sampled_networks_features, num_subsamples, subsample_size)

        for cur_subsam_idx in range(num_subsamples):
            print(f"{datetime.datetime.now()} [model_bootstrap] \t\t Working on subsample {cur_subsam_idx}/{num_subsamples}")
            sub_sample = sub_samples_features[:, cur_subsam_idx, :]
            sub_sample_mean = sub_sample.mean(axis=1)
            model_covariance_matrix = covariance_matrix_estimation(sub_sample, sub_sample_mean, method="naive")
            inv_model_cov_matrix = np.linalg.pinv(model_covariance_matrix)
            mahalanobis_dists[cur_subsam_idx] = mahalanobis(observed_features, sub_sample_mean, inv_model_cov_matrix)
        
        empirical_threshold = np.quantile(mahalanobis_dists, confidence)

        return {
            "success": empirical_threshold < stds_away_thr,
            "statistic": empirical_threshold,
            "threshold": stds_away_thr
        }


import unittest
from unittest.mock import patch, MagicMock

import numpy as np

from pyERGM.utils import *
from pyERGM.metrics import *

import networkx as nx
import math
import pandas as pd


class TestNumberOfEdgesUndirected(unittest.TestCase):
    def test_num_of_edges(self):
        metric = NumberOfEdgesUndirected()
        W = np.array([
            [0, 1, 1],
            [1, 0, 1],
            [1, 1, 0]
        ])

        result = metric.calculate(W)
        expected_result = 3

        self.assertEqual(result, expected_result)


class TestNumberOfEdgesDirected(unittest.TestCase):
    def test_num_of_edges(self):
        metric = NumberOfEdgesDirected()
        W = np.array([
            [0, 1, 0],
            [1, 0, 1],
            [1, 0, 0]
        ])

        result = metric.calculate(W)
        expected_result = 4

        self.assertEqual(result, expected_result)

    def test_num_of_edges_on_sample(self):
        metric = NumberOfEdgesDirected()
        n = 3
        sample_size = 2
        W1 = np.array([
            [0, 1, 0],
            [1, 0, 1],
            [1, 0, 0]
        ])
        W2 = np.array([
            [0, 0, 0],
            [1, 0, 1],
            [1, 0, 0]
        ])

        sample = np.zeros((n, n, sample_size))
        sample[:, :, 0] = W1
        sample[:, :, 1] = W2

        result = metric.calculate_for_sample(sample)
        expected_result = np.array([4, 3])

        self.assertTrue(np.all(result == expected_result))


class TestNumberOfTriangles(unittest.TestCase):
    def test_validation(self):
        metric = NumberOfTriangles()

        self.assertTrue(not metric.requires_graph)

        W = np.array([
            [0, 1, 0],
            [1, 0, 1],
            [1, 0, 0]
        ])

        with self.assertRaises(ValueError):
            metric.calculate(W)

    def test_num_of_triangles(self):
        metric = NumberOfTriangles()
        W = np.array([
            [0, 1, 1],
            [1, 0, 1],
            [1, 1, 0]
        ])

        result = metric.calculate(W)
        expected_result = 1

        self.assertEqual(result, expected_result)

        random_graph = nx.fast_gnp_random_graph(6, 0.5, directed=False, seed=42)
        triangles = nx.triangles(random_graph)
        total_num_triangles = sum(triangles.values()) / 3

        G_as_W = nx.to_numpy_array(random_graph)
        calculated_number_of_triangles = metric.calculate(G_as_W)

        self.assertEqual(total_num_triangles, calculated_number_of_triangles)

    def test_calc_change_score(self):
        metric = NumberOfTriangles()

        W = np.array([
            [0, 1, 1],
            [1, 0, 1],
            [1, 1, 0]
        ])

        idx_to_toggle = (0, 1)
        result = metric.calc_change_score(W, idx_to_toggle)
        self.assertEqual(result, -1)

        W = np.array([
            [0, 0, 0, 1],
            [0, 0, 1, 1],
            [0, 1, 0, 1],
            [1, 1, 1, 0]
        ])

        result = metric.calc_change_score(W, idx_to_toggle)
        self.assertEqual(result, 1)

        W = np.array([
            [0, 1, 0, 1],
            [1, 0, 1, 0],
            [0, 1, 0, 1],
            [1, 0, 1, 0]
        ])

        idx_to_toggle = (1, 3)
        result = metric.calc_change_score(W, idx_to_toggle)
        self.assertEqual(result, 2)

    def test_calculate_for_sample(self):
        np.random.seed(678)
        sample_size = 50
        n = 10
        networks_sample = np.zeros((n, n, sample_size))
        expected_triangles = np.zeros(sample_size)
        for i in range(sample_size):
            cur_graph = nx.fast_gnp_random_graph(n, 0.5, seed=np.random)
            expected_triangles[i] = sum(nx.triangles(cur_graph).values()) / 3
            networks_sample[:, :, i] = nx.to_numpy_array(cur_graph)

        res = NumberOfTriangles().calculate_for_sample(networks_sample)
        self.assertTrue(np.all(res == expected_triangles))


class TestReciprocity(unittest.TestCase):
    def test_calculate(self):
        metrics = [Reciprocity()]
        n = 3
        collection = MetricsCollection(metrics, is_directed=True, n_nodes=n)

        matrices_to_test = [
            {
                "W": np.array([
                    [0, 0, 1],
                    [1, 0, 0],
                    [0, 1, 0]
                ]),
                "expected_recip": np.array([0, 0, 0])
            },
            {
                "W": np.array([
                    [0, 0, 1],
                    [1, 0, 0],
                    [1, 1, 0]
                ]),
                "expected_recip": np.array([0, 1, 0])
            },
            {
                "W": np.array([
                    [0, 1, 1],
                    [1, 0, 1],
                    [1, 1, 0]
                ]),
                "expected_recip": np.array([1, 1, 1])
            }
        ]

        for test_data in matrices_to_test:
            W = test_data["W"]
            expected_reciprocity_vector = test_data["expected_recip"]
            g_y = collection.calculate_statistics(W)
            self.assertTrue(np.array_equal(g_y, expected_reciprocity_vector))

        n = 30
        collection = MetricsCollection(metrics, is_directed=True, n_nodes=n)
        W = np.random.randint(0, 2, (n, n))

        G = nx.from_numpy_array(W, create_using=nx.DiGraph)

        total_edges = np.sum(W)

        reciprocity_vector = collection.calculate_statistics(W)
        total_reciprocity = np.sum(reciprocity_vector)

        reciprocty_fraction = total_reciprocity / total_edges

        nx_reciprocity = nx.algorithms.reciprocity(G) / 2  # nx counts each reciprocity twice
        self.assertEqual(reciprocty_fraction, nx_reciprocity)


class TestDegreeMetrics(unittest.TestCase):
    def test_calculate(self):
        W = np.array([
            [0, 1, 0, 1],
            [1, 0, 0, 1],
            [1, 1, 0, 1],
            [1, 0, 0, 0]
        ])

        G = nx.from_numpy_array(W, create_using=nx.DiGraph)
        expected_in_degrees = list(dict(G.in_degree()).values())

        receiver = InDegree()
        indegree = receiver.calculate(W)

        self.assertTrue(all(indegree == expected_in_degrees))

        expected_out_degrees = list(dict(G.out_degree()).values())

        sender = OutDegree()
        outdegree = sender.calculate(W)

        self.assertTrue(all(outdegree == expected_out_degrees))

        undirected_degree = UndirectedDegree()
        W = np.array([
            [0, 1, 0, 1],
            [1, 0, 0, 1],
            [0, 0, 0, 1],
            [1, 1, 1, 0]
        ])

        G = nx.from_numpy_array(W)

        degrees = undirected_degree.calculate(W)
        expected_degrees = list(dict(G.degree()).values())
        self.assertTrue(all(degrees == expected_degrees))


class TestNumberOfEdgesTypesDirected(unittest.TestCase):
    def test_calc_edge_weights(self):
        neuronal_types = ['A', 'B', 'A', 'B']
        metric = NumberOfEdgesTypesDirected(neuronal_types)
        expected_edge_weights = np.array([[0, 1, 0, 1],
                                          [2, 3, 2, 3],
                                          [0, 1, 0, 1],
                                          [2, 3, 2, 3]])

        expected_edge_weights += 1  # Increment because we also want a bin for nonexisting edges (which have the entry 0)

        self.assertTrue(np.all(metric._edge_type_idx_assignment == expected_edge_weights))

        neuronal_types = ['A', 'B', 'B', 'A']
        metric = NumberOfEdgesTypesDirected(neuronal_types)
        expected_edge_weights = np.array([[0, 1, 1, 0],
                                          [2, 3, 3, 2],
                                          [2, 3, 3, 2],
                                          [0, 1, 1, 0]])

        expected_edge_weights += 1  # Increment because we also want a bin for nonexisting edges (which have the entry 0)

        print(metric._edge_type_idx_assignment)
        print(expected_edge_weights)
        self.assertTrue(np.all(metric._edge_type_idx_assignment == expected_edge_weights))

    def test_calculate(self):
        W = np.array([
            [0, 1, 0, 0],
            [1, 0, 0, 1],
            [1, 1, 0, 0],
            [0, 0, 1, 0]
        ])

        neuronal_types = ['A', 'B', 'A', 'B']
        expected_num_edges = np.array([1, 2, 2, 1])
        metric = NumberOfEdgesTypesDirected(neuronal_types)
        calculated_num_edges = metric.calculate(W)
        self.assertTrue(np.all(expected_num_edges == calculated_num_edges))

    def test_calculate_for_sample(self):
        n = 4
        sample_size = 2
        W1 = np.array([
            [0, 1, 0, 0],
            [1, 0, 0, 1],
            [1, 1, 0, 0],
            [0, 0, 1, 0]
        ])
        W2 = np.array([
            [0, 1, 1, 0],
            [0, 0, 0, 1],
            [1, 0, 0, 0],
            [0, 1, 1, 0]
        ])

        neuronal_types = ['A', 'B', 'A', 'B']
        expected_num_edges = np.array([
            [1, 2, 2, 1],
            [2, 1, 1, 2]
        ]).T

        sample = np.zeros((n, n, sample_size))
        sample[:, :, 0] = W1
        sample[:, :, 1] = W2
        metric = NumberOfEdgesTypesDirected(neuronal_types)
        calculated_num_edges = metric.calculate_for_sample(sample)
        self.assertTrue(np.all(expected_num_edges == calculated_num_edges))

    def test_calc_change_score(self):
        W1 = np.array([
            [0, 1, 0, 0],
            [1, 0, 0, 1],
            [1, 1, 0, 0],
            [0, 0, 1, 0]
        ])
        W2 = np.array([
            [0, 1, 1, 0],
            [0, 0, 0, 1],
            [1, 0, 0, 0],
            [0, 1, 1, 0]
        ])

        neuronal_types = ['A', 'B', 'A', 'B']
        metric = NumberOfEdgesTypesDirected(neuronal_types)

        calculated_change_score = metric.calc_change_score(W1, (0, 1))
        expected_change_score = np.array([0, -1, 0, 0])
        self.assertTrue(np.all(expected_change_score == calculated_change_score))

        calculated_change_score = metric.calc_change_score(W1, (2, 3))
        expected_change_score = np.array([0, 1, 0, 0])
        self.assertTrue(np.all(expected_change_score == calculated_change_score))

        calculated_change_score = metric.calc_change_score(W2, (0, 2))
        expected_change_score = np.array([-1, 0, 0, 0])
        self.assertTrue(np.all(expected_change_score == calculated_change_score))

        calculated_change_score = metric.calc_change_score(W2, (1, 2))
        expected_change_score = np.array([0, 0, 1, 0])
        self.assertTrue(np.all(expected_change_score == calculated_change_score))

    def test_calculate_change_score_full_network(self):
        types = ['A', 'B', 'A', 'B']
        metric = NumberOfEdgesTypesDirected(types)

        W1 = np.array([
            [0, 1, 0, 0],
            [1, 0, 0, 1],
            [1, 1, 0, 0],
            [0, 0, 1, 0]
        ])

        expected_full_change_score = np.array([
            [0, -1, 0, 0],
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, -1, 0],
            [0, 0, 1, 0],
            [0, 0, 0, -1],
            [-1, 0, 0, 0],
            [0, -1, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, 1],
            [0, 0, -1, 0]
        ])

        full_change_score = metric.calculate_change_score_full_network(W1)
        self.assertTrue(np.all(expected_full_change_score == full_change_score))


class TestNodeAttrSums(unittest.TestCase):
    def test_calc_edge_weights(self):
        node_attr = np.array([1, 2, 3, 4])
        metric_both = NodeAttrSum(node_attr, is_directed=True)
        expected_edge_weights = np.array([
            [0, 3, 4, 5],
            [3, 0, 5, 6],
            [4, 5, 0, 7],
            [5, 6, 7, 0]
        ])
        self.assertTrue(np.all(expected_edge_weights == metric_both.edge_weights))

        metric_out = NodeAttrSumOut(node_attr)
        expected_edge_weights = np.array([
            [0, 1, 1, 1],
            [2, 0, 2, 2],
            [3, 3, 0, 3],
            [4, 4, 4, 0]
        ])
        self.assertTrue(np.all(expected_edge_weights == metric_out.edge_weights))

        metric_in = NodeAttrSumIn(node_attr)
        self.assertTrue(np.all(expected_edge_weights.T == metric_in.edge_weights))

    def test_calculate(self):
        W = np.array([
            [0, 1, 1, 0],
            [0, 0, 0, 1],
            [1, 0, 0, 0],
            [0, 1, 1, 0]
        ])

        node_attr = np.array([1, 2, 3, 4])
        metric_both = NodeAttrSum(node_attr, is_directed=True)
        expected_statistic = 3 + 4 + 6 + 4 + 6 + 7
        self.assertEqual(expected_statistic, metric_both.calculate(W)[0])

        metric_out = NodeAttrSumOut(node_attr)
        expected_statistic = 1 + 1 + 2 + 3 + 4 + 4
        self.assertEqual(expected_statistic, metric_out.calculate(W)[0])

        metric_in = NodeAttrSumIn(node_attr)
        expected_statistic = 1 + 2 + 2 + 3 + 3 + 4
        self.assertEqual(expected_statistic, metric_in.calculate(W)[0])

    def test_calculate_for_sample(self):
        n = 3
        sample_size = 3

        W1 = np.array([
            [0, 1, 0],
            [1, 0, 1],
            [1, 0, 0]
        ])
        W2 = np.array([
            [0, 0, 1],
            [1, 0, 0],
            [1, 1, 0]
        ])
        W3 = np.array([
            [0, 0, 0],
            [1, 0, 0],
            [0, 1, 0]
        ])

        sample = np.zeros((n, n, sample_size))
        sample[:, :, 0] = W1
        sample[:, :, 1] = W2
        sample[:, :, 2] = W3

        node_attributes = np.array([2, 1, 1])

        metric_both = NodeAttrSum(node_attributes, is_directed=True)
        expected_stats_sample = np.array([11, 11, 5])
        calculated_stats_sample = metric_both.calculate_for_sample(sample)
        self.assertTrue(np.all(expected_stats_sample == calculated_stats_sample))

        metric_out = NodeAttrSumOut(node_attributes)
        expected_stats_sample = np.array([5, 5, 2])
        calculated_stats_sample = metric_out.calculate_for_sample(sample)
        self.assertTrue(np.all(expected_stats_sample == calculated_stats_sample))

        metric_in = NodeAttrSumIn(node_attributes)
        expected_stats_sample = np.array([6, 6, 3])
        calculated_stats_sample = metric_in.calculate_for_sample(sample)
        self.assertTrue(np.all(expected_stats_sample == calculated_stats_sample))

    def test_calc_change_score(self):
        W_off = np.zeros((3, 3))
        W_on = np.ones((3, 3))

        node_attributes = np.array([1, 0.5, 2])

        metric_both = NodeAttrSum(node_attributes, is_directed=True)
        metric_out = NodeAttrSumOut(node_attributes)
        metric_in = NodeAttrSumIn(node_attributes)

        calculated_change_score = metric_both.calc_change_score(W_off, (2, 1))
        expected_change_score = 2.5
        self.assertEqual(expected_change_score, calculated_change_score)

        calculated_change_score = metric_both.calc_change_score(W_on, (2, 0))
        expected_change_score = -3
        self.assertEqual(expected_change_score, calculated_change_score)

        calculated_change_score = metric_out.calc_change_score(W_off, (1, 0))
        expected_change_score = 0.5
        self.assertEqual(expected_change_score, calculated_change_score)

        calculated_change_score = metric_out.calc_change_score(W_on, (0, 2))
        expected_change_score = -1
        self.assertEqual(expected_change_score, calculated_change_score)

        calculated_change_score = metric_in.calc_change_score(W_off, (0, 1))
        expected_change_score = 0.5
        self.assertEqual(expected_change_score, calculated_change_score)

        calculated_change_score = metric_in.calc_change_score(W_on, (1, 2))
        expected_change_score = -2
        self.assertEqual(expected_change_score, calculated_change_score)

    def test_calculate_mple_regressors(self):
        # directed both (in+out)
        node_attr = np.array([3, 5, 7])
        W = np.array([[0, 1, 0],
                      [1, 0, 1],
                      [1, 0, 0]])

        attr_sum_mat = np.array([[0, 8, 10],
                                 [8, 0, 12],
                                 [10, 12, 0]])

        metric_both = NodeAttrSum(node_attr, is_directed=True)
        indices_lims = (0, 6)
        expected_regressors_1_1 = np.array([[8.],
                                            [10.],
                                            [8.],
                                            [12.],
                                            [10.],
                                            [12.]])
        self.assertTrue(np.all(metric_both.calculate_mple_regressors(W, indices_lims) == expected_regressors_1_1))

        indices_lims = (1, 4)
        expected_regressors_1_2 = np.array([[10.],
                                            [8.],
                                            [12.]])
        self.assertTrue(np.all(metric_both.calculate_mple_regressors(W, indices_lims) == expected_regressors_1_2))

        # directed in
        attr_sum_mat_in = np.array([[0, 5, 7],
                                    [3, 0, 7],
                                    [3, 5, 0]])

        metric_in = NodeAttrSumIn(node_attr)
        indices_lims = (0, 6)
        expected_regressors_2_1 = np.array([[5.],
                                            [7.],
                                            [3.],
                                            [7.],
                                            [3.],
                                            [5.]])
        self.assertTrue(np.all(metric_in.calculate_mple_regressors(W, indices_lims) == expected_regressors_2_1))

        indices_lims = (1, 4)
        expected_regressors_2_2 = np.array([[7.],
                                            [3.],
                                            [7.]])
        self.assertTrue(np.all(metric_in.calculate_mple_regressors(W, indices_lims) == expected_regressors_2_2))

        # directed out
        attr_sum_mat_out = np.array([[0, 3, 3],
                                     [5, 0, 5],
                                     [7, 7, 0]])

        metric_out = NodeAttrSumOut(node_attr)
        indices_lims = (0, 6)
        expected_regressors_3_1 = np.array([[3.],
                                            [3.],
                                            [5.],
                                            [5.],
                                            [7.],
                                            [7.]])

        self.assertTrue(np.all(metric_out.calculate_mple_regressors(W, indices_lims) == expected_regressors_3_1))

        indices_lims = (1, 4)
        expected_regressors_3_2 = np.array([[3.],
                                            [5.],
                                            [5.]])
        self.assertTrue(np.all(metric_out.calculate_mple_regressors(W, indices_lims) == expected_regressors_3_2))

        # undirected (both)
        node_attr = np.array([3, 5, 7])
        W = np.array([[0, 1, 0],
                      [1, 0, 1],
                      [0, 1, 0]])
        attr_sum_mat = np.array([[0, 8, 10],
                                 [8, 0, 12],
                                 [10, 12, 0]])

        metric_both = NodeAttrSum(node_attr, is_directed=False)
        indices_lims = (0, 3)
        # in undirected graphs the expected outcome is the upper triangle (without the diagonal)
        expected_regressors_1 = np.array([[8.],
                                          [10.],
                                          [12.]])
        self.assertTrue(np.all(metric_both.calculate_mple_regressors(W, indices_lims) == expected_regressors_1))

        indices_lims = (1, 2)
        expected_regressors_2 = np.array([[10.]])
        self.assertTrue(np.all(metric_both.calculate_mple_regressors(W, indices_lims) == expected_regressors_2))


class TestSumDistancesConnectedNeurons(unittest.TestCase):
    def test_sum_distances(self):
        positions = pd.DataFrame({"x_pos": [0, 0, 4], "y_pos": [0, 3, 0], "z_pos": [2, 2, 2]})
        W = np.array([[0, 1, 0],
                      [1, 0, 1],
                      [1, 0, 0]])

        # Dataframe with multiple columns
        metric_1 = SumDistancesConnectedNeurons(positions, is_directed=True)
        expected_res_1 = 3 + 3 + 5 + 4
        self.assertTrue(metric_1.calculate(W) == expected_res_1)

        # 2D numpy array
        metric_2 = SumDistancesConnectedNeurons(positions.to_numpy(), is_directed=True)
        expected_res_2 = 3 + 3 + 5 + 4
        self.assertTrue(metric_2.calculate(W) == expected_res_2)

        # Series
        metric_3 = SumDistancesConnectedNeurons(positions.x_pos, is_directed=True)
        expected_res_3 = 4 + 4
        self.assertTrue(metric_3.calculate(W) == expected_res_3)

        # 1D numpy array
        metric_4 = SumDistancesConnectedNeurons(positions.z_pos.to_numpy(), is_directed=True)
        expected_res_4 = 0
        self.assertTrue(metric_4.calculate(W) == expected_res_4)

        # list
        positions_y_list = [0, 3, 0]
        metric_5 = SumDistancesConnectedNeurons(positions_y_list, is_directed=True)
        expected_res_5 = 3 + 3 + 3
        self.assertTrue(metric_5.calculate(W) == expected_res_5)

        # tuple
        positions_y_tuple = (0, 3, 0)
        metric_6 = SumDistancesConnectedNeurons(positions_y_tuple, is_directed=True)
        expected_res_6 = 3 + 3 + 3
        self.assertTrue(metric_6.calculate(W) == expected_res_6)

        # undirected_graph
        W_undirected = np.array([[0, 1, 1],
                                 [1, 0, 0],
                                 [1, 0, 0]])

        metric_7 = SumDistancesConnectedNeurons(positions, is_directed=False)
        expected_res_7 = 3 + 4
        self.assertTrue(metric_7.calculate(W_undirected) == expected_res_7)

    def test_calculate_mple_regressors(self):
        positions = pd.DataFrame({"x_pos": [0, 0, 4], "y_pos": [0, 3, 0], "z_pos": [2, 2, 2]})
        W = np.array([[0, 1, 0],
                      [1, 0, 1],
                      [1, 0, 0]])

        # multiple columns
        metric_1 = SumDistancesConnectedNeurons(positions, is_directed=True)
        distances = np.array([[0, 3, 4],
                              [3, 0, 5],
                              [4, 5, 0]])
        indices_lims = (0, 6)
        expected_regressors_1_1 = np.array([[3.],
                                            [4.],
                                            [3.],
                                            [5.],
                                            [4.],
                                            [5.]])
        self.assertTrue(np.all(metric_1.calculate_mple_regressors(W, indices_lims) == expected_regressors_1_1))

        indices_lims = (1, 4)
        expected_regressors_1_2 = np.array([[4.],
                                            [3.],
                                            [5.]])
        self.assertTrue(np.all(metric_1.calculate_mple_regressors(W, indices_lims) == expected_regressors_1_2))

        # Series
        metric_2 = SumDistancesConnectedNeurons(positions.x_pos, is_directed=True)
        distances = np.array([[0, 0, 4],
                              [0, 0, 4],
                              [4, 4, 0]])
        indices_lims = (0, 6)
        expected_regressors_2_1 = np.array([[0.],
                                            [4.],
                                            [0.],
                                            [4.],
                                            [4.],
                                            [4.]])
        self.assertTrue(np.all(metric_2.calculate_mple_regressors(W, indices_lims) == expected_regressors_2_1))

        indices_lims = (1, 4)
        expected_regressors_2_2 = np.array([[4.],
                                            [0.],
                                            [4.]])
        self.assertTrue(np.all(metric_2.calculate_mple_regressors(W, indices_lims) == expected_regressors_2_2))

        metric_3 = SumDistancesConnectedNeurons(positions.z_pos, is_directed=True)
        distances = np.array([[0, 0, 0],
                              [0, 0, 0],
                              [0, 0, 0]])
        indices_lims = (0, 6)
        expected_regressors_3 = np.array([[0.],
                                          [0.],
                                          [0.],
                                          [0.],
                                          [0.],
                                          [0.]])
        self.assertTrue(np.all(metric_3.calculate_mple_regressors(W, indices_lims) == expected_regressors_3))

        # undirected
        metric_4 = SumDistancesConnectedNeurons(positions, is_directed=False)
        distances = np.array([[0, 3, 4],
                              [3, 0, 5],
                              [4, 5, 0]])
        indices_lims = (0, 3)
        expected_regressors_4_1 = np.array([[3.],
                                            [4.],
                                            [5.]])
        self.assertTrue(np.all(metric_4.calculate_mple_regressors(W, indices_lims) == expected_regressors_4_1))

        indices_lims = (1, 2)
        expected_regressors_4_2 = np.array([[4.]])
        self.assertTrue(np.all(metric_4.calculate_mple_regressors(W, indices_lims) == expected_regressors_4_2))

        metric_5 = SumDistancesConnectedNeurons(positions.z_pos, is_directed=False)
        distances = np.array([[0, 0, 0],
                              [0, 0, 0],
                              [0, 0, 0]])
        indices_lims = (0, 3)
        expected_regressors_5 = np.array([[0.],
                                          [0.],
                                          [0.]])
        self.assertTrue(np.all(metric_5.calculate_mple_regressors(W, indices_lims) == expected_regressors_5))


class TestNumberOfNodesPerType(unittest.TestCase):
    def test_calculate(self):
        V = np.array([[1, 0],
                      [2, 1],
                      [1, 1],
                      [0, 2]]) # n=4, k=2

        expected_num_neurons_per_type = np.array([1, 2, 1])
        metric = NumberOfNodesPerType(metric_node_feature={'morphology'}, n_node_categories=3)
        calculated_num_neurons_per_type = metric.calculate(V[:, [0]])
        self.assertTrue(np.all(expected_num_neurons_per_type[:-1] == calculated_num_neurons_per_type))

    def test_calculate_for_sample(self):
        n = 4
        sample_size = 2
        V1 = np.array([[1, 0],
                       [2, 1],
                       [1, 1],
                       [0, 2]])
        V2 = np.array([[2, 0],
                       [0, 1],
                       [0, 1],
                       [0, 1]])

        expected_num_neurons_per_type = np.array([
            [1, 2, 1],
            [3, 0, 1]
        ]).T

        sample = np.zeros((n, 1, sample_size), dtype=int)
        sample[:, :, 0] = V1[:, [0]]
        sample[:, :, 1] = V2[:, [0]]
        metric = NumberOfNodesPerType(metric_node_feature={'morphology'}, n_node_categories=3)
        calculated_num_neurons_per_type = metric.calculate_for_sample(sample)
        self.assertTrue(np.all(expected_num_neurons_per_type[:-1] == calculated_num_neurons_per_type))

    def test_calc_change_score(self):
        V1 = np.array([[1, 0],
                       [2, 1],
                       [1, 1],
                       [0, 2]])
        V2 = np.array([[2, 0],
                       [0, 1],
                       [0, 1],
                       [0, 1]])

        metric = NumberOfNodesPerType(metric_node_feature={'morphology'}, n_node_categories=3)

        calculated_change_score = metric.calc_change_score(V1[:, [0]], idx=2, new_category=0)
        expected_change_score = np.array([1, -1, 0])
        self.assertTrue(np.all(expected_change_score[:-1] == calculated_change_score))

        calculated_change_score = metric.calc_change_score(V1[:, [1]], idx=0, new_category=2)
        expected_change_score = np.array([-1, 0, 1])
        self.assertTrue(np.all(expected_change_score[:-1] == calculated_change_score))

        calculated_change_score = metric.calc_change_score(V2[:, [0]], idx=1, new_category=1)
        expected_change_score = np.array([-1, 1, 0])
        self.assertTrue(np.all(expected_change_score[:-1] == calculated_change_score))

        calculated_change_score = metric.calc_change_score(V2[:, [1]], idx=0, new_category=1)
        expected_change_score = np.array([-1, 1, 0])
        self.assertTrue(np.all(expected_change_score[:-1] == calculated_change_score))


class TestMetricsCollection(unittest.TestCase):

    def test_get_effective_feature_count(self):
        n = 18
        receiver = InDegree()

        collection = MetricsCollection([receiver], is_directed=True, n_nodes=n, do_copy_metrics=False)
        self.assertEqual(receiver._get_effective_feature_count(), n)

        receiver = InDegree(indices_from_user=[0])
        collection = MetricsCollection([receiver], is_directed=True, n_nodes=n, do_copy_metrics=False)
        self.assertEqual(receiver._get_effective_feature_count(), n - 1)

        sender = OutDegree()

        collection = MetricsCollection([sender], is_directed=True, n_nodes=n, do_copy_metrics=False)
        self.assertEqual(sender._get_effective_feature_count(), n)

        sender = OutDegree(indices_from_user=[0])
        collection = MetricsCollection([sender], is_directed=True, n_nodes=n, do_copy_metrics=False)
        self.assertEqual(sender._get_effective_feature_count(), n - 1)

    def test_metrics_setup(self):
        metrics = [NumberOfEdgesDirected(), TotalReciprocity(), InDegree()]
        collection = MetricsCollection(metrics, is_directed=True, n_nodes=3)

        expected_metric_names = tuple([str(NumberOfEdgesDirected()), str(TotalReciprocity()), str(InDegree())])

        self.assertTrue(collection.metric_names == expected_metric_names)

        metrics = [InDegree(), UndirectedDegree()]
        with self.assertRaises(ValueError):
            collection = MetricsCollection(metrics, is_directed=True,
                                           n_nodes=3)  # Should fail because we have an undirected metric in a directed metrics collection

        with self.assertRaises(ValueError):
            collection = MetricsCollection(metrics, is_directed=False,
                                           n_nodes=3)  # Should fail because we have a directed metric in an undirected metrics collection

    def test_collinearity_fixer(self):
        n = 18

        possible_types = ['A', 'B', 'C', 'D']
        # sampled once using np.random.choice(len(possible_types), size=n)
        type_idx_per_node = [3, 3, 0, 0, 0, 0, 1, 3, 1, 1, 3, 0, 3, 2, 0, 3, 3, 1]
        neuronal_types = [possible_types[type_idx_per_node[i]] for i in range(n)]

        n_for_multitypes_test = 6
        multitypes_1 = ['A', 'B']
        multitypes_2 = ['C', 'D']
        type_1_idx_per_node = [0, 0, 0, 1, 1, 1]
        type_2_idx_per_node = [0, 0, 0, 0, 1, 1]
        types_1 = [multitypes_1[type_1_idx_per_node[i]] for i in range(n_for_multitypes_test)]
        types_2 = [multitypes_2[type_2_idx_per_node[i]] for i in range(n_for_multitypes_test)]

        multitypes_3 = ['A', 'B']
        multitypes_4 = ['C', 'D', 'E']
        type_3_idx_per_node = [0, 0, 1, 1, 1, 1]
        type_4_idx_per_node = [0, 0, 1, 1, 2, 2]
        types_3 = [multitypes_3[type_3_idx_per_node[i]] for i in range(n_for_multitypes_test)]
        types_4 = [multitypes_4[type_4_idx_per_node[i]] for i in range(n_for_multitypes_test)]

        test_scenarios = {
            "num_edges_twice": {
                "metrics": [NumberOfEdgesDirected(), NumberOfEdgesDirected()],
                "expected_num_of_features": 1,
                "expected_trimmed_metrics": {},
                "expected_eliminated_metrics": []
            },
            "num_edges__total_reciprocity__indegree": {
                "metrics": [NumberOfEdgesDirected(), TotalReciprocity(), InDegree()],
                "expected_num_of_features": 1 + 1 + n - 1,
                "expected_trimmed_metrics": {},
                "expected_eliminated_metrics": []
            },
            "indegree": {
                "metrics": [InDegree()],
                "expected_num_of_features": n,  # There is no collinearity since InDegree() is the only feature
                "expected_trimmed_metrics": {},
                "expected_eliminated_metrics": []
            },
            "outdegree": {
                "metrics": [OutDegree()],
                "expected_num_of_features": n,
                "expected_trimmed_metrics": {},
                "expected_eliminated_metrics": []
            },
            "in_outdegree": {
                "metrics": [InDegree(), OutDegree()],
                "expected_num_of_features": n + n - 1,
                "expected_trimmed_metrics": {},
                "expected_eliminated_metrics": []
            },
            "num_edges__total_reciprocity__indegree_outdegree": {
                "metrics": [NumberOfEdgesDirected(), TotalReciprocity(), InDegree(), OutDegree()],
                "expected_num_of_features": 1 + 1 + n - 1 + n - 1,
                "expected_trimmed_metrics": {},
                "expected_eliminated_metrics": []
            },
            "num_edges__exogenous_types": {
                "metrics": [NumberOfEdgesDirected(), NumberOfEdgesTypesDirected(neuronal_types)],
                "expected_num_of_features": 1 + len(set(neuronal_types)) ** 2 - 2,
                "expected_trimmed_metrics": {},
                "expected_eliminated_metrics": []
            },
            "sum_attr_both__sum_attr_in__sum_attr_out": {
                "metrics": [NodeAttrSum(np.arange(n), is_directed=True), NodeAttrSumIn(np.arange(n)),
                            NodeAttrSumOut(np.arange(n))],
                "expected_num_of_features": 2,
                "expected_trimmed_metrics": {},
                "expected_eliminated_metrics": []
            },
            "multiple_types": {
                "n": n_for_multitypes_test,
                "metrics": [NumberOfEdgesDirected(), NumberOfEdgesTypesDirected(types_1),
                            NumberOfEdgesTypesDirected(types_2)],
                "expected_num_of_features": 1 + len(set(types_1)) ** 2 + len(set(types_2)) ** 2 - 2,
                "expected_trimmed_metrics": {},
                "expected_eliminated_metrics": []
            },
            "multiple_types_2": {
                "n": n_for_multitypes_test,
                "metrics": [NumberOfEdgesDirected(), NumberOfEdgesTypesDirected(types_3),
                            NumberOfEdgesTypesDirected(types_4)],
                "expected_num_of_features": 1 + len(set(types_3)) ** 2 + len(set(types_4)) ** 2 - 5,
                "expected_trimmed_metrics": {},
                "expected_eliminated_metrics": []
            }
        }

        # We want to make sure that the changes (trimming and complete removal) are made for the right metrics, so we
        # update the scenarios after initializing the lists of Metrics, to pass the class instances by reference when
        # asserting stuff later.
        test_scenarios["num_edges_twice"]["expected_eliminated_metrics"] = [
            test_scenarios["num_edges_twice"]["metrics"][0]]

        test_scenarios["num_edges__total_reciprocity__indegree"]["expected_trimmed_metrics"] = {
            # InDegree is trimmed by 1 because of collinearity with num_of_edges
            test_scenarios["num_edges__total_reciprocity__indegree"]["metrics"][2]: [0]
        }

        test_scenarios["in_outdegree"]["expected_trimmed_metrics"] = {
            test_scenarios["in_outdegree"]["metrics"][0]: [0]
        }

        test_scenarios["num_edges__total_reciprocity__indegree_outdegree"]["expected_trimmed_metrics"] = {
            test_scenarios["num_edges__total_reciprocity__indegree_outdegree"]["metrics"][2]: [0],
            test_scenarios["num_edges__total_reciprocity__indegree_outdegree"]["metrics"][3]: [0]
        }

        test_scenarios["num_edges__exogenous_types"]["expected_trimmed_metrics"] = {
            test_scenarios["num_edges__exogenous_types"]["metrics"][1]: [0, 10]
        }

        test_scenarios["sum_attr_both__sum_attr_in__sum_attr_out"]["expected_eliminated_metrics"] = [
            test_scenarios["sum_attr_both__sum_attr_in__sum_attr_out"]["metrics"][0]]

        test_scenarios["multiple_types"]["expected_trimmed_metrics"] = {
            test_scenarios["multiple_types"]["metrics"][1]: [0],
            test_scenarios["multiple_types"]["metrics"][2]: [0]}

        test_scenarios["multiple_types_2"]["expected_trimmed_metrics"] = {
            # A->A, A->B, B->A trimmed
            test_scenarios["multiple_types_2"]["metrics"][1]: [0, 1, 2],
            # C->C, D->D trimmed
            test_scenarios["multiple_types_2"]["metrics"][2]: [0, 4]}

        for name, scenario_data in test_scenarios.items():
            net_size = scenario_data.get("n", n)

            collection = MetricsCollection(scenario_data["metrics"], is_directed=True, fix_collinearity=True,
                                           n_nodes=net_size, do_copy_metrics=False)

            # Check the general number of features matches the expectation
            self.assertEqual(collection.num_of_features, scenario_data["expected_num_of_features"])

            # Check if the correct metrics were trimmed
            for metric in scenario_data["metrics"]:
                if metric in scenario_data["expected_trimmed_metrics"].keys():
                    self.assertEqual(set(np.where(metric._indices_to_ignore)[0]),
                                     set(scenario_data["expected_trimmed_metrics"][metric]))
                elif hasattr(metric,
                             "_indices_to_ignore"):  # Only metrics with a base_idx can be trimmed. The others aren't tested
                    self.assertFalse(np.any(metric._indices_to_ignore))
                elif metric in scenario_data["expected_eliminated_metrics"]:
                    self.assertTrue(metric not in [m for m in collection.metrics])
                    self.assertEqual(collection.num_of_metrics, len(collection.metrics))

    def test_calculate_statistics(self):
        # Test undirected graphs
        metrics = [NumberOfEdgesUndirected(), NumberOfTriangles()]
        collection = MetricsCollection(metrics, is_directed=False, n_nodes=3)

        W = np.array([
            [0, 1, 1],
            [1, 0, 1],
            [1, 1, 0]
        ])

        stats = collection.calculate_statistics(W)
        expected_stats = np.array([3, 1])

        np.testing.assert_array_equal(stats, expected_stats)

        # Test directed graphs
        metrics = [NumberOfEdgesDirected()]
        collection = MetricsCollection(metrics, is_directed=True, n_nodes=3)
        W = np.array([
            [0, 1, 0],
            [1, 0, 1],
            [1, 0, 0]
        ])

        stats = collection.calculate_statistics(W)
        expected_stats = np.array([4])

        np.testing.assert_array_equal(stats, expected_stats)

    def test_calculate_statistics_with_node_features(self):
        # Test undirected graphs
        metrics = [NumberOfEdgesUndirected(), NumberOfTriangles()]
        collection = MetricsCollection(metrics, is_directed=False, n_nodes=3)

        # Now, W is an (n, n+k) with n=3, k=2. Result shouldn't change.
        W = np.array([
            [0, 1, 1, 1, 2],
            [1, 0, 1, 0, 1],
            [1, 1, 0, 1, 0]
        ])

        stats = collection.calculate_statistics(W)
        expected_stats = np.array([3, 1])

        np.testing.assert_array_equal(stats, expected_stats)

        # Test directed graphs
        metrics = [NumberOfEdgesDirected()]
        collection = MetricsCollection(metrics, is_directed=True, n_nodes=3)
        W = np.array([
            [0, 1, 0],
            [1, 0, 1],
            [1, 0, 0]
        ])

        stats = collection.calculate_statistics(W)
        expected_stats = np.array([4])

        np.testing.assert_array_equal(stats, expected_stats)

    def test_get_num_of_features(self):
        n = 4
        metrics = [NumberOfEdgesUndirected(), NumberOfTriangles()]
        collection = MetricsCollection(metrics, is_directed=False, n_nodes=n)
        num_of_features = collection.num_of_features

        # NumberOfEdges and NumberOfTriangles each produce 1 features
        expected_num_of_features = 2

        self.assertEqual(num_of_features, expected_num_of_features)

        metrics = [Reciprocity(), NumberOfEdgesDirected()]
        collection = MetricsCollection(metrics, is_directed=True, n_nodes=n)

        num_of_features = collection.num_of_features

        expected_num_of_features = math.comb(n, 2) + 1

        self.assertEqual(num_of_features, expected_num_of_features)

    def test_calc_change_scores(self):
        metrics = [NumberOfEdgesDirected(), Reciprocity()]
        collection = MetricsCollection(metrics, is_directed=True, n_nodes=3)

        W1 = np.array([
            [0, 0, 1],
            [1, 0, 0],
            [1, 1, 0]
        ])

        W2 = np.array([
            [0, 0, 1],
            [1, 0, 0],
            [0, 1, 0]
        ])

        flipped_indices = (2, 0)
        edge_flip_info = {
            'edge': flipped_indices
        }
        result = collection.calc_change_scores(W1, edge_flip_info=edge_flip_info)

        # 1st is -1 because we lost an edge, and 3rd entry is -1 because node #2 lost it's reciprocity
        expected_result = [-1, 0, -1, 0]

        self.assertTrue(all(result == expected_result))

    def test_calculate_change_scores_all_edges(self):
        types = ['A', 'B', 'A', 'B']
        metrics = [NumberOfEdgesDirected(), NumberOfEdgesTypesDirected(types)]

        W1 = np.array([
            [0, 1, 0, 0],
            [1, 0, 0, 1],
            [1, 1, 0, 0],
            [0, 0, 1, 0]
        ])
        n_nodes = W1.shape[0]

        expected_full_change_score = np.array([
            [-1, 0, -1, 0, 0],
            [1, 1, 0, 0, 0],
            [1, 0, 1, 0, 0],
            [-1, 0, 0, -1, 0],
            [1, 0, 0, 1, 0],
            [-1, 0, 0, 0, -1],
            [-1, -1, 0, 0, 0],
            [-1, 0, -1, 0, 0],
            [1, 0, 1, 0, 0],
            [1, 0, 0, 1, 0],
            [1, 0, 0, 0, 1],
            [-1, 0, 0, -1, 0]
        ])

        collection = MetricsCollection(metrics, is_directed=True, n_nodes=n_nodes)

        # The collinearity fixer is supposed to remove one attribute from the NumberOfEdgesTypesDirected metric
        # self.assertEqual(len(collection.metrics[1]._indices_to_ignore), 1)
        self.assertEqual(np.sum(collection.metrics[1]._indices_to_ignore), 1)

        idx_to_ignore = np.where(collection.metrics[1]._indices_to_ignore)[0][0]

        res = collection.calculate_change_scores_all_edges(W1)

        # Deleting the 1+idx_to_ignore because the first entry is the NumberOfEdgesDirected metric
        expected_full_change_score = np.delete(expected_full_change_score, 1 + idx_to_ignore, axis=1)
        self.assertTrue(np.all(expected_full_change_score == res))

    def test_prepare_mple_data(self):
        types = ['A', 'B', 'A', 'B']
        metrics = [NumberOfEdgesDirected(), NumberOfEdgesTypesDirected(types)]

        W1 = np.array([
            [0, 1, 0, 0],
            [1, 0, 0, 1],
            [1, 1, 0, 0],
            [0, 0, 1, 0]
        ])

        n_nodes = W1.shape[0]
        expected_flattened_mat = np.array([1, 0, 0, 1, 0, 1, 1, 1, 0, 0, 0, 1]).reshape(n_nodes ** 2 - n_nodes, 1)

        expected_mple_regressors = np.array([
            [1, 0, 1, 0, 0],
            [1, 1, 0, 0, 0],
            [1, 0, 1, 0, 0],
            [1, 0, 0, 1, 0],
            [1, 0, 0, 1, 0],
            [1, 0, 0, 0, 1],
            [1, 1, 0, 0, 0],
            [1, 0, 1, 0, 0],
            [1, 0, 1, 0, 0],
            [1, 0, 0, 1, 0],
            [1, 0, 0, 0, 1],
            [1, 0, 0, 1, 0]
        ])

        collection = MetricsCollection(metrics, is_directed=True, n_nodes=n_nodes)

        # The collinearity fixer is supposed to remove one attribute from the NumberOfEdgesTypesDirected metric
        self.assertEqual(np.sum(collection.metrics[1]._indices_to_ignore), 1)

        idx_to_ignore = np.where(collection.metrics[1]._indices_to_ignore)[0][0]

        Xs_full, ys_full = collection.prepare_mple_data(W1)

        Xs_half, ys_half = collection.prepare_mple_data(W1,
                                                        edges_indices_lims=(0, expected_mple_regressors.shape[0] // 2))

        # Deleting the 1+idx_to_ignore because the first entry is the NumberOfEdgesDirected metric
        expected_mple_regressors = np.delete(expected_mple_regressors, 1 + idx_to_ignore, axis=1)
        self.assertTrue(np.all(expected_mple_regressors == Xs_full))
        self.assertTrue(np.all(expected_mple_regressors[:expected_mple_regressors.shape[0] // 2] == Xs_half))
        self.assertTrue(np.all(expected_flattened_mat == ys_full))
        self.assertTrue(np.all(expected_flattened_mat[:expected_mple_regressors.shape[0] // 2] == ys_half))

    def test_get_parameter_names(self):
        n = 18
        metrics = [NumberOfEdgesDirected(), InDegree(), OutDegree()]
        collection = MetricsCollection(metrics, is_directed=True, n_nodes=n)

        param_names = collection.get_parameter_names()

        expected_names = ()
        for metric in metrics:
            metric_name = str(metric)
            if metric_name == "num_edges_directed":
                expected_names += (metric_name,)
            elif metric_name == "indegree" or metric_name == "outdegree":
                for i in range(1, n):
                    expected_names += (f"{metric_name}_{i + 1}",)

        self.assertTrue(param_names == expected_names)

        n = 4
        metrics = [NumberOfEdgesTypesDirected(['A', 'B', 'A', 'B'])]
        collection = MetricsCollection(metrics, is_directed=True, n_nodes=n)

        param_names = collection.get_parameter_names()

        expected_names = ("num_edges_between_types_directed_A__A", "num_edges_between_types_directed_A__B",
                          "num_edges_between_types_directed_B__A", "num_edges_between_types_directed_B__B")
        self.assertTrue(param_names == expected_names)

    def test_get_ignored_features(self):
        n = 18
        metrics = [NumberOfEdgesDirected(), InDegree(), OutDegree()]
        collection = MetricsCollection(metrics, is_directed=True, n_nodes=n)

        ignored_features = collection.get_ignored_features()
        expected_ignored_features = ("indegree_1", "outdegree_1")
        self.assertTrue(ignored_features == expected_ignored_features)

    @patch('pyERGM.metrics.split_network_for_bootstrapping')
    def test_bootstrapped_features(self, mock_split_network_for_bootstrapping):
        W = np.array([
            [0, 1, 0, 0],
            [1, 0, 0, 1],
            [1, 0, 0, 1],
            [0, 1, 1, 0]
        ])

        W_symmetric = np.array([
            [0, 1, 0, 0],
            [1, 0, 0, 1],
            [0, 0, 0, 1],
            [0, 1, 1, 0]
        ])

        mock_split_network_for_bootstrapping.return_value = (
            np.array([0, 2]).reshape((2, 1)), np.array([1, 3]).reshape((2, 1)))

        metrics = [NumberOfEdgesDirected()]
        metrics_collection = MetricsCollection(metrics, is_directed=True, n_nodes=W.shape[0])
        bootstrapped_features = metrics_collection.bootstrap_observed_features(W, 1)
        expected_bootstrapped_features = np.array([6]).reshape(1, 1)  # np.sum(W[[0,2],[0,2].T]) / 2 * 12
        self.assertTrue(np.all(bootstrapped_features == expected_bootstrapped_features))

        metrics = [NumberOfEdgesUndirected()]
        metrics_collection = MetricsCollection(metrics, is_directed=False, n_nodes=W_symmetric.shape[0])
        bootstrapped_features = metrics_collection.bootstrap_observed_features(W_symmetric, 1)
        expected_bootstrapped_features = np.array([0]).reshape(1, 1)  # nodes 0 and 2 are not connected
        self.assertTrue(np.all(bootstrapped_features == expected_bootstrapped_features))

        metrics = [TotalReciprocity()]
        metrics_collection = MetricsCollection(metrics, is_directed=True, n_nodes=W.shape[0])
        bootstrapped_features = metrics_collection.bootstrap_observed_features(W, 1)
        expected_bootstrapped_features = np.array([0]).reshape(1, 1)
        self.assertTrue(np.all(bootstrapped_features == expected_bootstrapped_features))

        metrics = [UndirectedDegree()]
        metrics_collection = MetricsCollection(metrics, is_directed=False, n_nodes=W_symmetric.shape[0])
        bootstrapped_features = metrics_collection.bootstrap_observed_features(W_symmetric, 1)
        expected_bootstrapped_features = np.array([0, 3, 0, 3]).reshape(4,
                                                                        1)  # nodes 0,2 are not connected, and nodes 1,3 are.
        self.assertTrue(np.all(bootstrapped_features == expected_bootstrapped_features))

        metrics = [InDegree()]
        metrics_collection = MetricsCollection(metrics, is_directed=True, n_nodes=W.shape[0])
        bootstrapped_features = metrics_collection.bootstrap_observed_features(W, 1)
        expected_bootstrapped_features = np.array([3, 3, 0, 3]).reshape(4,
                                                                        1)  # e.g., the first is given by np.sum(W[[0,2],[0,2].T], axis=0) / (2-1) * (4-1)
        self.assertTrue(np.all(bootstrapped_features == expected_bootstrapped_features))

        metrics = [NumberOfEdgesDirected(), OutDegree()]
        metrics_collection = MetricsCollection(metrics, is_directed=True, n_nodes=W.shape[0])
        bootstrapped_features = metrics_collection.bootstrap_observed_features(W, 1)
        expected_bootstrapped_features = np.array([6, 3, 3, 3]).reshape(4,
                                                                        1)  # Ignoring the out-degree of the first node
        self.assertTrue(np.all(bootstrapped_features == expected_bootstrapped_features))

        mock_split_network_for_bootstrapping.return_value = (
            np.array([2, 3]).reshape((2, 1)), np.array([0, 1]).reshape((2, 1)))
        metrics = [TotalReciprocity()]
        metrics_collection = MetricsCollection(metrics, is_directed=True, n_nodes=W.shape[0])
        bootstrapped_features = metrics_collection.bootstrap_observed_features(W, 1)
        expected_bootstrapped_features = np.array([6]).reshape(1, 1)
        self.assertTrue(np.all(bootstrapped_features == expected_bootstrapped_features))



from setuptools import setup, find_packages

setup(
    name="pyERGM",
    version="0.2.0",
    description="Exponential Random Graph Models in Python",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/tomtalp/pyERGM",
    packages=find_packages(),
    python_requires=">=3.10",
)


import itertools
from collections import Counter
from typing import Collection
import numpy as np
import networkx as nx
from numba import njit, objmode
from scipy.spatial.distance import mahalanobis
from scipy.optimize import minimize, OptimizeResult
import torch
import random
import time
from pathlib import Path
import os
import pickle
import shutil
import subprocess
import sys

# Indices of columns in the output of bjobs -A
JOB_ARRAY_ID_IDX = 0
ARRAY_SPEC_IDX = 1
DONE_IDX = 5
EXIT_IDX = 7

LSF_ID_LIST_LEN_LIMIT = 100


def perturb_network_by_overriding_edge(network, value, i, j, is_directed):
    perturbed_net = network.copy()
    perturbed_net[i, j] = value
    if not is_directed:
        perturbed_net[j, i] = value

    return perturbed_net


def connectivity_matrix_to_G(W: np.ndarray, directed=False):
    """
    Convert a connectivity matrix to a graph object.
    
    Parameters
    ----------
    W : np.ndarray
        A connectivity matrix.
        
    Returns
    -------
    G : nx.Graph
        A graph object.

    """
    if directed:
        G = nx.from_numpy_array(W, create_using=nx.DiGraph)
    else:
        G = nx.from_numpy_array(W)
    return G


def get_random_nondiagonal_matrix_entry(n: int):
    """
    Get a random entry for a non-diagonal entry of a matrix.
    
    Parameters
    ----------
    n : int
        The size of the matrix.
        
    Returns
    -------
    entry : tuple
        The row and column index of the entry.
    """
    xs = list(range(n))
    return tuple(random.sample(xs, 2))


def construct_adj_mat_from_int(int_code: int, num_nodes: int, is_directed: bool) -> np.ndarray:
    num_pos_connects = num_nodes * (num_nodes - 1)
    if not is_directed:
        num_pos_connects //= 2
    adj_mat_str = f'{int_code:0{num_pos_connects}b}'
    cur_adj_mat = np.zeros((num_nodes, num_nodes))
    mat_entries_arr = np.array(list(adj_mat_str), 'uint8')
    if is_directed:
        cur_adj_mat[~np.eye(num_nodes, dtype=bool)] = mat_entries_arr
    else:
        upper_triangle_indices = np.triu_indices(num_nodes, k=1)
        cur_adj_mat[upper_triangle_indices] = mat_entries_arr
        lower_triangle_indices_aligned = (upper_triangle_indices[1], upper_triangle_indices[0])
        cur_adj_mat[lower_triangle_indices_aligned] = mat_entries_arr
    return cur_adj_mat


def construct_int_from_adj_mat(adj_mat: np.ndarray, is_directed: bool) -> int:
    if len(adj_mat.shape) != 2 or adj_mat.shape[0] != adj_mat.shape[1]:
        raise ValueError(f"The dimensions of the given adjacency matrix {adj_mat.shape} are not valid for an "
                         f"adjacency matrix (should be a 2D squared matrix)")
    num_nodes = adj_mat.shape[0]
    adj_mat_no_diag = adj_mat[~np.eye(num_nodes, dtype=bool)]
    if not is_directed:
        adj_mat_no_diag = adj_mat_no_diag[:adj_mat_no_diag.size // 2]
    return round((adj_mat_no_diag * 2 ** np.arange(adj_mat_no_diag.size - 1, -1, -1).astype(np.ulonglong)).sum())


def _calc_line_slope_2_points(xs: np.ndarray, ys: np.ndarray):
    if xs.size != 2 or ys.size != 2:
        raise ValueError(
            f"The sizes of xs and ys must be 2 to form a well defined line! Got {xs.size}, {ys.size} instead")
    return (xs)


def get_greatest_convex_minorant(xs: np.ndarray, ys: np.ndarray):
    if xs.size != ys.size:
        raise ValueError("Arrays must have the same size!")
    # For numerical considerations we don't force the sequence of slopes to really not decrease
    slope_diff_thr = -10 ** -10
    x_diffs = np.diff(xs)
    # TODO: this case be handled (reduce the larger y value to the lower one, perform the same logic for the effective
    #  n-1 points, and duplicate the relevant point at the end). But is seems a very specific case and I'm not sure it's
    #  worth the complication.
    if np.any(x_diffs == 0):
        raise ValueError("xs array must contain unique values!")
    cur_proposed_minorant = ys.copy()
    cur_spline_slopes = np.diff(cur_proposed_minorant) / x_diffs
    cur_decreasing_slope_indices = np.where(np.diff(cur_spline_slopes) < slope_diff_thr)[0]
    # Convexity condition - the slopes are monotonically increasing
    while cur_decreasing_slope_indices.size > 0:
        # The indices of the points that violate the condition, and whose y values need to be decreases are in 1 shift
        # from the problematic slopes (if the second slope is smaller than the first slope, cur_decreasing_slope_indices
        # would include the index 0, of the first slopes difference. This corresponds to the point that takes part in
        # the calculations of both slopes, the first and the second, which is the second point, with index 1).
        cur_problematic_indices = cur_decreasing_slope_indices + 1

        # These are the indices that wrap the problematic ones (i.e., all the indices that are between wrapping_indices
        # are problematic).
        # Note: problematic indices may be subsequent (indices 2 and 3 are problematic, so we want 1, 4 to wrap them).
        cur_wrapping_indices = np.array([i for i in range(xs.size) if (i not in cur_problematic_indices and (
                i + 1 in cur_problematic_indices or i - 1 in cur_problematic_indices))]).astype(int)

        # To get an equation of a line from the slope and a point on in we can rearrange the formula:
        # y-y0 = m(x-x0) -> y = mx + (y0-m*x0)
        cur_wrapping_lines_slopes = np.diff(cur_proposed_minorant[cur_wrapping_indices]) / np.diff(
            xs[cur_wrapping_indices])
        cur_wrapping_lines_intersepts = (cur_proposed_minorant[cur_wrapping_indices[:-1]] -
                                         cur_wrapping_lines_slopes * xs[cur_wrapping_indices[:-1]])

        # Update the values in problematic indices to lie on the corresponding wrapping line
        cur_assignment_problematic_idx_to_line = np.searchsorted(cur_wrapping_indices, cur_problematic_indices) - 1
        cur_proposed_minorant[cur_problematic_indices] = (
                cur_wrapping_lines_slopes[cur_assignment_problematic_idx_to_line] * xs[cur_problematic_indices] +
                cur_wrapping_lines_intersepts[cur_assignment_problematic_idx_to_line])

        # Update the slopes and the indices where they decrease
        cur_spline_slopes = np.diff(cur_proposed_minorant) / x_diffs
        cur_decreasing_slope_indices = np.where(np.diff(cur_spline_slopes) < slope_diff_thr)[0]

    return cur_proposed_minorant


@njit
def get_uniform_random_edges_to_flip(num_nodes, num_pairs):
    """
    Create a matrix of size (2 x num_pairs), where each column represents a pair of nodes.
    These nodes represent the edge we wish to flip.
    The pairs are sampled randomly.
    """

    edges_to_flip = np.zeros((2, num_pairs), dtype=np.int32)

    edges_to_flip[0, :] = np.random.choice(num_nodes, size=num_pairs)

    diff = np.random.choice(num_nodes - 1, size=num_pairs) + 1

    edges_to_flip[1, :] = (edges_to_flip[0, :] - diff) % num_nodes

    return edges_to_flip


def get_uniform_random_nodes_to_flip(num_nodes, num_flips):
    """
    Create a vector of size num_flips, where each entry represents a node we wish to flip.
    The nodes are sampled randomly.
    """

    nodes_to_flip = np.random.choice(num_nodes, size=num_flips)

    return nodes_to_flip

def get_uniform_random_new_node_feature_categories(node_features_to_flip, node_features_inds_to_n_categories):
    """
    Create a dictionary of all the possible category flips for the predetermined node features.
    keys are node features indices, and values are dictionaries with the key being a category, and the value being
    random new categories (excluding the current category) - a vector of size of the number of appearances of the
    feature in the random feature flips array given as input. Each entry represents the new category to flip to.
    Totally, we save num_flips x (mean_num_categories - 1) numbers.
    The categories are sampled randomly.
    """

    new_node_features_categories = {}
    for feature_ind, n_categories in node_features_inds_to_n_categories.items():
        categories = np.arange(n_categories)
        new_node_feature_categories = {}
        for c in categories:
            categories_without_c = np.delete(categories, c)
            random_new_categories = np.random.choice(categories_without_c, size=(node_features_to_flip == feature_ind).sum())
            new_node_feature_categories[c] = random_new_categories
        new_node_features_categories[feature_ind] = new_node_feature_categories
    return new_node_features_categories


def convert_flat_no_diag_idx_to_i_j(flat_no_diag_idx: Collection[int], full_mat_size: int) -> np.ndarray[int]:
    """
    Converts the index in the flattened square matrix without the main diagonal to the pair of indices in the original
    matrix.
    For a square matrix A, the flattened-no-diagonal form is given by A[~np.eye(A.shape[0], dtype=bool)].
    E.g., given a 3X3 matrix, the third element in its flattened-no-diagonal form (idx 2) is the (1,0) entry.
    Parameters
    ----------
    flat_no_diag_idx
        The index in the flattened-no-diagonal form
        #TODO: force this to be a np.ndarray and use numba?
    full_mat_size
        The number of rows/columns in the original squared matrix.
    Returns
    -------
    The tuple of indices of the entry in the original square matrix.
    """
    flat_no_diag_idx = np.array(flat_no_diag_idx)
    if np.any(flat_no_diag_idx >= full_mat_size * (full_mat_size - 1)):
        raise IndexError(
            f"Got a too large `flat_no_diag_idx` {flat_no_diag_idx} for original matrix size of {full_mat_size}")
    rows = np.array(flat_no_diag_idx) // (full_mat_size - 1)
    cols = np.array(flat_no_diag_idx) % (full_mat_size - 1)
    cols[cols >= rows] += 1
    return np.stack((rows, cols), axis=0).astype(np.int32)


def get_custom_distribution_random_edges_to_flip(num_pairs, edge_probs):
    """
    Sample pairs of indices for edge flips according to a given probability distribution over all possible edges (all
    entries of the adjacency matrix but the main diagonal).
    Parameters
    ----------
    num_pairs
        The sample size
    edge_probs
        The probability distribution over all possible edges. The indexing here is in the flattened-no-diagonal format.
        This is equivalent to having a matrix A of nXn probabilities, where the i,j-th entry is the probability of the
        i,j-th edge to be sampled, and passing A[~np.eye(A.shape[0], dtype=bool)].
    Returns
    -------
    A sample of num_pairs pairs of indices, sampled according to edge_probs.
    """
    num_possible_edges = edge_probs.size
    # The solution for the equation n(n-1)=num_possible_edges
    num_nodes = np.round((1 + np.sqrt(1 + 4 * num_possible_edges)) / 2)  # TODO: validate that this is indeed an int?
    flat_no_diag_indices = np.random.choice(edge_probs.size, p=edge_probs, size=num_pairs)
    # TODO: force the following to be pre-complied with numba, and the current as well?
    return convert_flat_no_diag_idx_to_i_j(flat_no_diag_indices, num_nodes)


def np_tensor_to_sparse_tensor(np_tensor: np.ndarray) -> torch.Tensor:
    """
    Receives a numpy tensor and converts it to a sparse Tensor, using Torch.
    TODO - Support different types of Sparse Matrix data structures? More efficient conversion?
    """
    return torch.from_numpy(np_tensor).to_sparse()


def transpose_sparse_sample_matrices(sparse_tensor: torch.Tensor) -> torch.Tensor:
    """
    Transpose a sparse tensor that represents k matrices of dimension n x n.
    The transpose operation occurs along the dimension of sample (i.e. each matrix is transposed separately)

    Parameters
    ----------
    sparse_tensor: torch.Tensor
        A sparse tensor of dimension (n, n, k) representing k matrices of dim (n,n)

    Returns
    -------
    transposed_tensor: torch.Tensor
        The same tensor but matrices are transposed

    """
    n = sparse_tensor.shape[0]
    k = sparse_tensor.shape[2]

    indices = sparse_tensor.indices().type(torch.int64)
    transposed_indices = torch.stack([indices[1], indices[0], indices[2]])
    values = sparse_tensor.values()

    return torch.sparse_coo_tensor(transposed_indices, values, (n, n, k))


def calc_for_sample_njit():
    def wrapper(func):
        def inner(sample):
            if isinstance(sample, np.ndarray):
                return njit(func)(sample)
            return func(sample)

        return inner

    return wrapper


def approximate_auto_correlation_function(features_of_net_samples: np.ndarray) -> np.ndarray:
    """
    This is gamma hat from Geyer's handbook of mcmc (1D) and Dai and Jones 2017 (multi-D).
    """
    # TODO: it must be possible to vectorize this calculation and spare the for loop. Maybe somehow use the
    #  convolution theorem and go back and forth to the frequency domain using FFT for calculating correlations.
    features_mean_diff = features_of_net_samples - features_of_net_samples.mean(axis=1)[:, None]
    num_features = features_of_net_samples.shape[0]
    sample_size = features_of_net_samples.shape[1]
    auto_correlation_func = np.zeros((sample_size, num_features, num_features))
    for k in range(sample_size):
        auto_correlation_func[k] = 1 / sample_size * (
                features_mean_diff[:, :sample_size - k].T.reshape(sample_size - k, num_features, 1) @
                features_mean_diff[:, k:].T.reshape(sample_size - k, 1, num_features)
        ).sum(axis=0)
    return auto_correlation_func


@njit
def approximate_kth_auto_correlation_function(features_mean_diff: np.ndarray, lag: int) -> np.ndarray:
    num_features = features_mean_diff.shape[0]
    sample_size = features_mean_diff.shape[1]
    head = features_mean_diff[:, :sample_size - lag]
    tail = features_mean_diff[:, lag:]
    mean_feat_diff_outer_prods = np.zeros((num_features, num_features, sample_size - lag))
    for i in range(sample_size - lag):
        mean_feat_diff_outer_prods[:, :, i] = np.outer(head[:, i], tail[:, i])
    return 1 / sample_size * mean_feat_diff_outer_prods.sum(axis=2)


def calc_capital_gammas(auto_corr_funcs: np.ndarray) -> np.ndarray:
    """
    This is the capital gammas hat from Geyer's handbook of mcmc (1D) and Dai and Jones 2017 (multi-D).
    They are simply summations over pairs of consecutive even and odd indices of the auto correlation function (gammas).
    """
    # From Dai and Jones 2017 - a mean of gamma with its transpose (which corresponds to the negative index with the
    # same abs value).
    gamma_tilde = (auto_corr_funcs + np.transpose(auto_corr_funcs, [0, 2, 1])) / 2

    # Note - we assume here an even sample_size, it is forced elsewhere (everytime the sample size is updated).
    sample_size = gamma_tilde.shape[0]
    return (gamma_tilde[np.arange(0, sample_size - 1, 2, dtype=int)] +
            gamma_tilde[np.arange(1, sample_size, 2, dtype=int)])


@njit
def calc_kth_capital_gamma(features_mean_diff: np.ndarray, k: int) -> np.ndarray:
    two_k_gamma_hat = approximate_kth_auto_correlation_function(features_mean_diff, 2 * k)
    two_k_plus_one_gamma_hat = approximate_kth_auto_correlation_function(features_mean_diff, 2 * k + 1)
    return (two_k_gamma_hat + two_k_gamma_hat.T + two_k_plus_one_gamma_hat + two_k_plus_one_gamma_hat.T) / 2


@njit
def covariance_matrix_estimation(features_of_net_samples: np.ndarray, mean_features_of_net_samples: np.ndarray,
                                 method='batch', num_batches=25) -> np.ndarray:
    """
    Approximate the covariance matrix of the model's features
    Parameters
    ----------
    features_of_net_samples
        The calculated features of the networks that are used for the approximation. Of dimensions
        (num_features X sample_size)
    mean_features_of_net_samples
        The mean of the features across the samples (a vector of size num_features)
    method
        the method to use for approximating the covariance matrix
        currently supported options are:
            naive
                A naive estimation from the sample: E[gi*gj] - E[gi]E[gj]
            batch
                based on difference of means of sample batches from the total mean, as in Geyer's handbook of
                MCMC (there it is stated for the univariate case, but the generalization is straight forward).
            multivariate_initial_sequence
                Following Dai and Jones 2017 - the first estimator in section 3.1 (denoted mIS).
    TODO: implement a mechanism that allows to pass arguments that are customized for each method

    Returns
    -------
    The covariance matrix estimation (num_features X num_features).
    """
    num_features = features_of_net_samples.shape[0]
    sample_size = features_of_net_samples.shape[1]
    if method == 'naive':
        # An outer product of the means (E[gi]E[gj])
        cross_prod_mean_features = (mean_features_of_net_samples.reshape(num_features, 1) @
                                    mean_features_of_net_samples.T.reshape(1, num_features))
        # A mean of the outer products of the sample (E[gi*gj])
        features_cross_prods = np.zeros((sample_size, num_features, num_features))
        for i in range(sample_size):
            features_cross_prods[i] = np.outer(features_of_net_samples[:, i], features_of_net_samples[:, i])
        mean_features_cross_prod = features_cross_prods.sum(axis=0) / sample_size

        return mean_features_cross_prod - cross_prod_mean_features

    elif method == 'batch':
        # Verify that the sample is nicely divided into non-overlapping batches.
        while sample_size % num_batches != 0:
            num_batches += 1
        batch_size = sample_size // num_batches

        # Divide the sample into batches, and calculate the mean of each one of them
        batches_means = np.zeros((num_features, num_batches))
        for i in range(num_batches):
            batches_means[:, i] = features_of_net_samples[:, i * batch_size:(i + 1) * batch_size].sum(
                axis=1) / batch_size

        diff_of_global_mean = batches_means - mean_features_of_net_samples.reshape(num_features, 1)

        # Average the outer products of the differences between batch means and the global mean and multiply by the
        # batch size to compensate for the aggregation into batches
        diff_of_global_mean_outer_prods = np.zeros((num_batches, num_features, num_features))
        for i in range(num_batches):
            diff_of_global_mean_outer_prods[i] = np.outer(diff_of_global_mean[:, i], diff_of_global_mean[:, i])
        batches_cov_mat_est = batch_size * diff_of_global_mean_outer_prods.sum(axis=0) / num_batches

        return batches_cov_mat_est

    elif method == "multivariate_initial_sequence":
        features_mean_diff = features_of_net_samples - mean_features_of_net_samples.reshape(num_features, 1)

        # In this method, we sum up capital gammas, and choose where to cut the tail (which corresponds to estimates
        # of auto-correlations with large lags within the chain. Naturally, as the lag increases the estimation
        # becomes worse, so the magic here is to determine where to cut). So we calculate the estimates one by one and
        # evaluate the conditions for cutting.
        # The first condition is to have an index where the estimation is positive-definite, namely all eigen-values
        # are positive. As both gamma_0 (which is auto_corr_funcs[0]) and the capital gammas are symmetric, all
        # the sum of them is allways symmetric, which ensures real eigen values, and we can simply calculate the
        # eigen value with the smallest algebraic value to determine whether all of them are positive.
        is_positive = False
        first_pos_def_idx = 0
        cur_pos_cov_mat_est = np.zeros((num_features, num_features))
        gamma_hat_0 = approximate_kth_auto_correlation_function(features_mean_diff, 0)
        while not is_positive:
            if first_pos_def_idx == sample_size // 2:
                # TODO: ValueError? probably should throw something else. And maybe it is better to try alone some
                #  of the remediations suggested here and just notify the user...
                raise ValueError("Got a sample with no valid multivariate_initial_sequence covariance matrix "
                                 "estimation (no possibility is positive-definite). Consider increasing sample size"
                                 " or using a different covariance matrix estimation method.")
            cur_capital_gamma = calc_kth_capital_gamma(features_mean_diff, first_pos_def_idx)
            if first_pos_def_idx == 0:
                cur_pos_cov_mat_est = -gamma_hat_0 + 2 * cur_capital_gamma
            else:
                cur_pos_cov_mat_est += 2 * cur_capital_gamma
            cur_smallest_eigen_val = np.linalg.eigvalsh(cur_pos_cov_mat_est)[0]
            if cur_smallest_eigen_val > 0:
                is_positive = True
            else:
                first_pos_def_idx += 1

        # Now we find the farthest idx after first_pos_def_idx for which the sequence of determinants is strictly
        # monotonically increasing.
        do_dets_increase = True
        cutting_idx = first_pos_def_idx
        cur_det = np.linalg.det(cur_pos_cov_mat_est)
        while do_dets_increase and cutting_idx < sample_size // 2:
            cur_capital_gamma = calc_kth_capital_gamma(features_mean_diff, cutting_idx + 1)
            next_pos_cov_mat_est = cur_pos_cov_mat_est + 2 * cur_capital_gamma
            next_det = np.linalg.det(next_pos_cov_mat_est)
            if next_det <= cur_det:
                do_dets_increase = False
            else:
                cutting_idx += 1
                cur_pos_cov_mat_est = next_pos_cov_mat_est
                cur_det = next_det

        return cur_pos_cov_mat_est

    else:
        raise ValueError(f"{method} is an unsupported method for covariance matrix estimation")


@njit
def calc_nll_gradient(observed_features, mean_features_of_net_samples):
    return mean_features_of_net_samples - observed_features


def get_sorted_type_pairs(types):
    sorted_types = sorted(list(set(types)))
    return list(itertools.product(sorted_types, sorted_types))


def get_edge_density_per_type_pairs(W: np.ndarray, types: Collection):
    """
    Calculate the density of edges between each pair of types in the network.

    Parameters
    ----------
    W : np.ndarray
        The adjacency matrix of the network
    
    types : Collection
        A list of types for every node in a network

    Returns
    ----------
    densities : np.ndarray
        An array of edge density per type pairs, which is calculated as follows -    
        p_(type1, type2) = (number of edges between type1 and type2) / (number of potential edges between type1 and type2).

        Array size is k^2 where k is number of types
    """
    type_pairs = get_sorted_type_pairs(types)
    n = W.shape[0]
    real_frequencies = {k: 0 for k in type_pairs}

    types_frequencies = dict(Counter(types))
    potential_frequencies = {}

    # Count how many potential edges can exist between each pair of types
    for pair in type_pairs:
        type_1 = pair[0]
        type_2 = pair[1]

        if type_1 == type_2:
            potential_frequencies[pair] = types_frequencies[type_1] * (types_frequencies[type_1] - 1)
        else:
            potential_frequencies[pair] = types_frequencies[type_1] * types_frequencies[type_2]

    # Count how many actual edges exist between each pair of types
    for i in range(n):
        for j in range(n):
            if i == j:
                continue

            type_a = types[i]
            type_b = types[j]

            real_frequencies[(type_a, type_b)] += W[i, j]

    normalized_real_frequencies = {k: 0 if potential_frequencies[k] == 0 else v / potential_frequencies[k] for k, v in
                                   real_frequencies.items()}
    return normalized_real_frequencies


def calc_hotelling_statistic_for_sample(observed_features: np.ndarray, sample_features: np.ndarray,
                                        cov_mat_est_method: str):
    mean_features = np.mean(sample_features, axis=1)
    cov_mat_est = covariance_matrix_estimation(sample_features, mean_features,
                                               method=cov_mat_est_method)
    inv_cov_mat = np.linalg.pinv(cov_mat_est)
    dist = mahalanobis(observed_features, mean_features, inv_cov_mat)
    sample_size = sample_features.shape[1]
    hotelling_t_stat = sample_size * dist * dist
    num_features = sample_features.shape[0]
    hotelling_t_as_f = ((sample_size - num_features) / (
            num_features * (sample_size - 1))) * hotelling_t_stat
    return hotelling_t_as_f


@njit
def sigmoid(x: np.ndarray | float):
    return 1 / (1 + np.exp(-x))


@njit
def calc_logistic_regression_predictions(Xs: np.ndarray, thetas: np.ndarray):
    """
    Calculate the predictions of a Logistic Regression model with input Xs and parameters thetas
    Parameters
    ----------
    Xs
        The input to the model (the regressors), of shape (num_samples X num_features)
    thetas
        The parameters of the model, of shape (num_features X 1)
    Returns
        sigmoid(Xs @ thetas)
    -------
    """
    return sigmoid(Xs @ thetas)


@njit
def calc_logistic_regression_predictions_log_likelihood(predictions: np.ndarray, ys: np.ndarray, eps=1e-10):
    """
    Calculates the log-likelihood of labeled data with regard to the predictions of a Logistic Regression model.
    Parameters
    ----------
    predictions
        The predictions of a Logistic Regression model (the probabilities it assigns to feature vectors to be
        labeled 1). Of shape (num_samples X 1)
    ys
        The labeled data (zeros and ones, for each feature vector). Of shape (num_samples X 1)
    Returns
    -------
    The probability to observe the vector `ys` under the distribution induced by `predictions`.
    """
    trimmed_likelihoods = np.clip(predictions, eps, 1 - eps)
    data_zero_indices = np.where(ys == 0)[0]
    trimmed_likelihoods[data_zero_indices] = np.ones((data_zero_indices.size, 1)) - trimmed_likelihoods[
        data_zero_indices]

    return np.log(trimmed_likelihoods).sum()


@njit
def calc_logistic_regression_log_likelihood_grad(Xs: np.ndarray, predictions: np.ndarray, ys: np.ndarray):
    """
    Calculates the gradient of the log-likelihood of labeled data with regard to the predictions of a Logistic
    Regression model.
    Parameters
    ----------
    Xs
        The input to the model (the regressors), of shape (num_samples X num_features)
    predictions
        The predictions of a Logistic Regression model (the probabilities it assigns to feature vectors to be
        labeled 1). Of shape (num_samples X 1)
    ys
         The labeled data (zeros and ones, for each feature vector). Of shape (num_samples X 1)
    Returns
    -------
    The gradient (partial derivatives with relation to thetas - the model parameters) of the log-likelihood of the data
    given the model. Of shape (num_features X 1)
    """
    return Xs.T @ (ys - predictions)


@njit
def calc_logistic_regression_log_likelihood_hessian(Xs: np.ndarray, predictions: np.ndarray):
    """
    Calculates the hessian of the log-likelihood of labeled data with regard to the predictions of a Logistic
    Regression model.
    Parameters
    ----------
    Xs
        The input to the model (the regressors), of shape (num_samples X num_features)
    predictions
        The predictions of a Logistic Regression model (the probabilities it assigns to feature vectors to be
        labeled 1). Of shape (num_samples X 1)
    Returns
    -------
    The hessian (partial second derivatives with relation to thetas - the model parameters) of the log-likelihood of the
    data given the model. Of shape (num_features X num_features)
    """
    return Xs.T @ (predictions * (1 - predictions) * Xs)


def calc_logistic_regression_log_likelihood_from_x_thetas(Xs: np.ndarray, thetas: np.ndarray, ys: np.ndarray):
    """
    Calculates the log-likelihood of labeled data with regard to the predictions of a Logistic Regression model.
    Parameters
    ----------
    Xs
        The input to the model (the regressors), of shape (num_samples X num_features)
    thetas
        The parameters of the model, of shape (num_features X 1)
    ys
        The labeled data (zeros and ones, for each feature vector). Of shape (num_samples X 1)
    Returns
    -------
    The probability to observe the vector `ys` under the distribution induced by the model.
    """
    return (-np.log(1 + np.exp(-Xs @ thetas)) + (ys - 1) * Xs @ thetas).sum()


def analytical_minus_log_likelihood_local(thetas, Xs, ys, eps=1e-10):
    pred = np.clip(calc_logistic_regression_predictions(Xs, thetas.reshape(thetas.size, 1)), eps, 1 - eps)
    return -calc_logistic_regression_predictions_log_likelihood(pred, ys)


@njit
def numerically_stable_minus_log_like_and_grad_local(thetas, Xs, ys):
    linear_pred = Xs @ thetas

    # Magic numbers are taken from sklearn:
    # https://github.com/scikit-learn/scikit-learn/blob/72b35a46684c0ecf4182500d3320836607d1f17c/sklearn/_loss/_loss.pyx.tp#L728
    rng_1_idx = np.where(linear_pred <= -37)[0]
    rng_2_idx = np.where((-37 < linear_pred) & (linear_pred <= -2))[0]
    rng_3_idx = np.where((-2 < linear_pred) & (linear_pred <= 18))[0]
    rng_4_idx = np.where(linear_pred > 18)[0]

    minus_log_like = 0
    minus_log_like_der_per_sample = np.zeros(linear_pred.size)
    if rng_1_idx.size > 0:
        exp_rng_1 = np.exp(linear_pred[rng_1_idx])
        minus_log_like += (exp_rng_1 - ys[rng_1_idx, 0] * linear_pred[rng_1_idx]).sum()
        minus_log_like_der_per_sample[:rng_1_idx.size] = exp_rng_1 - ys[rng_1_idx, 0]
    if rng_2_idx.size > 0:
        exp_rng_2 = np.exp(linear_pred[rng_2_idx])
        minus_log_like += (
                np.log(1 + exp_rng_2) - ys[rng_2_idx, 0] * linear_pred[rng_2_idx]).sum()
        minus_log_like_der_per_sample[rng_1_idx.size:rng_1_idx.size + rng_2_idx.size] = ((1 - ys[
            rng_2_idx, 0]) * exp_rng_2 - ys[rng_2_idx, 0]) / (1 + exp_rng_2)
    if rng_3_idx.size > 0:
        exp_rng_3 = np.exp(-linear_pred[rng_3_idx])
        minus_log_like += (
                np.log(1 + exp_rng_3) + (1 - ys[rng_3_idx, 0]) * linear_pred[rng_3_idx]).sum()
        minus_log_like_der_per_sample[
        rng_1_idx.size + rng_2_idx.size: minus_log_like_der_per_sample.size - rng_4_idx.size
        ] = ((1 - ys[rng_3_idx, 0]) - ys[rng_3_idx, 0] * exp_rng_3) / (1 + exp_rng_3)
    if rng_4_idx.size > 0:
        exp_rng_4 = np.exp(-linear_pred[rng_4_idx])
        minus_log_like += (exp_rng_4 + (1 - ys[rng_4_idx, 0]) * linear_pred[rng_4_idx]).sum()
        minus_log_like_der_per_sample[minus_log_like_der_per_sample.size - rng_4_idx.size:
        ] = ((1 - ys[rng_4_idx, 0]) - ys[rng_4_idx, 0] * exp_rng_4) / (1 + exp_rng_4)

    return minus_log_like, Xs.T @ minus_log_like_der_per_sample


def analytical_minus_log_likelihood_distributed(thetas, data_path, num_edges_per_job):
    return -distributed_logistic_regression_optimization_step(data_path, thetas.reshape(thetas.size, 1),
                                                              'log_likelihood', num_edges_per_job)


def analytical_logistic_regression_predictions_distributed(thetas, data_path, num_edges_per_job):
    return distributed_logistic_regression_optimization_step(data_path, thetas.reshape(thetas.size, 1),
                                                             'predictions', num_edges_per_job)


def analytical_minus_log_like_grad_local(thetas, Xs, ys, eps=1e-10):
    pred = np.clip(calc_logistic_regression_predictions(Xs, thetas.reshape(thetas.size, 1)), eps, 1 - eps)
    return -calc_logistic_regression_log_likelihood_grad(Xs, pred, ys).reshape(thetas.size, )


def analytical_minus_log_like_grad_distributed(thetas, data_path, num_edges_per_job):
    return -distributed_logistic_regression_optimization_step(data_path, thetas.reshape(thetas.size, 1),
                                                              'log_likelihood_gradient',
                                                              num_edges_per_job).reshape(thetas.size, )


def analytical_minus_log_likelihood_hessian_local(thetas, Xs, ys, eps=1e-10):
    pred = np.clip(calc_logistic_regression_predictions(Xs, thetas.reshape(thetas.size, 1)), eps, 1 - eps)
    return -calc_logistic_regression_log_likelihood_hessian(Xs, pred)


def analytical_minus_log_likelihood_hessian_distributed(thetas, data_path, num_edges_per_job):
    return -distributed_logistic_regression_optimization_step(data_path, thetas.reshape(thetas.size, 1),
                                                              'log_likelihood_hessian',
                                                              num_edges_per_job)


def mple_logistic_regression_optimization(metrics_collection, observed_network: np.ndarray,
                                          initial_thetas: np.ndarray | None = None,
                                          is_distributed: bool = False, optimization_method: str = 'L-BFGS-B',
                                          **kwargs):
    """
    Optimize the parameters of a Logistic Regression model by maximizing the likelihood using scipy.optimize.minimize.
    Parameters
    ----------
    metrics_collection
        The `MetricsCollection` with relation to which the optimization is carried out.
        # TODO: we can't add a type hint for this, due to circular import (utils can't import from metrics, as metrics
            already imports from utils). This might suggest that this isn't the right place for this function.
    observed_network
        The observed network used as data for the optimization.
    initial_thetas
        The initial vector of parameters. If `None`, the initial state is randomly sampled from (0,1)
    is_distributed
        Whether the calculations are carried locally or distributed over many compute nodes of an IBM LSF cluster.
    optimization_method
        The optimization method to use. Currently only 'L-BFGS-B' and 'Newton-CG' are supported.
    num_edges_per_job
        The number of graph edges (representing data points in this optimization) to consider for each job. Relevant
        only for distributed optimization.
    
    Returns
    -------
    thetas: np.ndarray
        The optimized parameters of the model
    pred: np.ndarray
        The predictions of the model on the observed network
    success: bool
        Whether the optimization was successful
    """

    def after_iteration_callback(intermediate_result: OptimizeResult):
        nonlocal iteration
        iteration += 1
        cur_time = time.time()
        print(f'iteration: {iteration}, time from start '
              f'training: {cur_time - start_time} '
              f'log10 likelihood: {-intermediate_result.fun / np.log(10)}')
        sys.stdout.flush()

    iteration = 0
    start_time = time.time()
    print("optimization started")
    sys.stdout.flush()

    if not is_distributed:
        Xs, ys = metrics_collection.prepare_mple_data(observed_network)
    else:
        out_dir_path = (Path.cwd().parent / "OptimizationIntermediateCalculations").resolve()
        data_path = (out_dir_path / "data").resolve()
        os.makedirs(data_path, exist_ok=True)

        # Copy the `MetricsCollection` and the observed network to provide its path to children jobs, so they will be
        # able to access it.
        metric_collection_path = os.path.join(data_path, 'metric_collection.pkl')
        with open(metric_collection_path, 'wb') as f:
            pickle.dump(metrics_collection, f)
        observed_net_path = os.path.join(data_path, 'observed_network.pkl')
        with open(observed_net_path, 'wb') as f:
            pickle.dump(observed_network, f)

    num_features = metrics_collection.calc_num_of_features()
    if initial_thetas is None:
        thetas = np.random.rand(num_features)
    else:
        thetas = initial_thetas.copy()

    if not is_distributed:
        if optimization_method == "Newton-CG":
            res = minimize(analytical_minus_log_likelihood_local, thetas, args=(Xs, ys),
                           jac=analytical_minus_log_like_grad_local, hess=analytical_minus_log_likelihood_hessian_local,
                           callback=after_iteration_callback, method="Newton-CG")
        elif optimization_method == "L-BFGS-B":
            res = minimize(analytical_minus_log_likelihood_local, thetas, args=(Xs, ys), 
                           jac=analytical_minus_log_like_grad_local, method="L-BFGS-B", callback=after_iteration_callback)
        else:
            raise ValueError(
                f"Unsupported optimization method: {optimization_method}. Options are: Newton-CG, L-BFGS-B")
        pred = calc_logistic_regression_predictions(Xs, res.x.reshape(-1, 1)).flatten()
    else:
        # TODO: support L-BFGS-B as an optimization method, and implement a numerically stable version for distributed
        #  calculations as well.
        num_edges_per_job = kwargs.get('num_edges_per_job', 100000)
        if optimization_method == "Newton-CG":
            res = minimize(analytical_minus_log_likelihood_distributed, thetas, args=(data_path, num_edges_per_job),
                           jac=analytical_minus_log_like_grad_distributed,
                           hess=analytical_minus_log_likelihood_hessian_distributed,
                           callback=after_iteration_callback, method="Newton-CG")
        else:
            raise ValueError(
                f"Unsupported optimization method: {optimization_method} for distributed optimization. "
                f"Options are: Newton-CG")
        pred = analytical_logistic_regression_predictions_distributed(res.x.reshape(-1, 1), data_path,
                                                                      num_edges_per_job).flatten()

    print(res)
    sys.stdout.flush()
    return res.x, pred, res.success


def distributed_logistic_regression_optimization_step(data_path, thetas, func_to_calc, num_edges_per_job=5000):
    # TODO: support calculating multiple functions in a single job array, maybe by getting an array of function
    #  names rather than a single function as func_to_calc. This will enable passing jac=True to scipy.optimize.minimize
    #  when the optimization method is L-BFGS-B, and calculating both log-likelihood and gradient in a single job array,
    #  reducing the number of sent jobs significantly.

    # Arrange files and send the children jobs
    num_jobs, out_path, job_array_ids = _run_distributed_logistic_regression_children_jobs(data_path, thetas,
                                                                                           func_to_calc,
                                                                                           num_edges_per_job)

    # Wait for all jobs to finish.
    chunks_path = (out_path / func_to_calc).resolve()
    os.makedirs(chunks_path, exist_ok=True)
    wait_for_distributed_children_outputs(num_jobs, chunks_path, job_array_ids, func_to_calc)
    # Clean current scripts
    shutil.rmtree((out_path / "scripts").resolve())

    # Aggregate results
    if func_to_calc == "predictions":
        aggregated_func = cat_children_jobs_outputs(num_jobs, (out_path / func_to_calc).resolve())
    else:
        aggregated_func = _sum_children_jobs_outputs(num_jobs, (out_path / func_to_calc).resolve())

    # Clean current outputs
    shutil.rmtree((out_path / func_to_calc).resolve())
    return aggregated_func


def _construct_single_batch_bash_cmd_logistic_regression(out_path, cur_thetas, func_to_calculate, num_edges_per_job):
    # Construct a string with the current thetas, to pass using the command line to children jobs.
    thetas_str = ''
    for t in cur_thetas:
        thetas_str += f'{t[0]} '
    thetas_str = thetas_str[:-1]

    cmd_line_for_bsub = (f'python ./logistic_regression_distributed_calcs.py '
                         f'--out_dir_path {out_path} '
                         f'--num_edges_per_job {num_edges_per_job} '
                         f'--function {func_to_calculate} '
                         f'--thetas {thetas_str}')
    return cmd_line_for_bsub


def run_distributed_children_jobs(out_path, cmd_line_single_batch, single_batch_template_file_name, num_jobs,
                                  array_name):
    # Create current bash scripts to send distributed calculations
    scripts_path = (out_path / "scripts").resolve()
    os.makedirs(scripts_path, exist_ok=True)
    single_batch_bash_path = os.path.join(scripts_path, "single_batch.sh")
    shutil.copyfile(os.path.join(os.getcwd(), "ClusterScripts", single_batch_template_file_name),
                    single_batch_bash_path)
    with open(single_batch_bash_path, 'a') as f:
        f.write(cmd_line_single_batch)

    multiple_batches_bash_path = os.path.join(scripts_path, "multiple_batches.sh")
    with open(multiple_batches_bash_path, 'w') as f:
        num_rows = 1
        while (num_rows - 1) * 2000 < num_jobs:
            f.write(f'bsub < $1 -J {array_name}'
                    f'[{(num_rows - 1) * 2000 + 1}-{min(num_rows * 2000, num_jobs)}]\n')
            num_rows += 1

    # Make sure the logs directory for the children jobs exists and delete previous logs.
    with open(single_batch_bash_path, 'r') as f:
        single_batch_bash_txt = f.read()
    logs_rel_dir_start = single_batch_bash_txt.find('-o') + 3
    log_rel_dir_end = single_batch_bash_txt.find('outs.%J.%I.log')
    logs_rel_dir = single_batch_bash_txt[logs_rel_dir_start:log_rel_dir_end]
    logs_dir = os.path.join(os.getcwd(), logs_rel_dir)
    os.makedirs(logs_dir, exist_ok=True)
    for file_name in os.listdir(logs_dir):
        os.unlink(os.path.join(logs_dir, file_name))

    # Send the jobs
    send_jobs_command = f'bash {multiple_batches_bash_path} {single_batch_bash_path}'
    jobs_sending_res = subprocess.run(send_jobs_command.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    job_array_ids = _parse_sent_job_array_ids(jobs_sending_res.stdout)

    return job_array_ids


def _run_distributed_logistic_regression_children_jobs(data_path, cur_thetas, func_to_calculate, num_edges_per_job):
    out_path = data_path.parent

    cmd_line_single_batch = _construct_single_batch_bash_cmd_logistic_regression(out_path, cur_thetas,
                                                                                 func_to_calculate, num_edges_per_job)

    with open(os.path.join(data_path, "observed_network.pkl"), 'rb') as f:
        observed_network = pickle.load(f)

    num_nodes = observed_network.shape[0]
    num_data_points = num_nodes * num_nodes - num_nodes
    num_jobs = int(np.ceil(num_data_points / num_edges_per_job))

    job_array_ids = run_distributed_children_jobs(out_path, cmd_line_single_batch, "distributed_logistic_regression.sh",
                                                  num_jobs, func_to_calculate)
    return num_jobs, out_path, job_array_ids


def wait_for_distributed_children_outputs(num_jobs: int, out_path: Path, job_array_ids: list, array_name: str):
    is_all_done = False
    time_of_last_job_status_check = time.time()
    num_jobs_to_listen_to = num_jobs
    are_there_missing_jobs = True
    # Wait that all output files will be there (all jobs finished successfully).
    while are_there_missing_jobs:
        # Wait for the current running jobs to finish
        while not is_all_done:
            if time.time() - time_of_last_job_status_check > 60:
                is_all_done = _check_if_all_done(job_array_ids, num_jobs_to_listen_to)
                time_of_last_job_status_check = time.time()
        print("is_all_done is True")
        sys.stdout.flush()

        # Reset the ids of arrays and number of jobs to listen to, as previous ones are done.
        job_array_ids = []
        num_jobs_to_listen_to = 0

        # Identify missing output files
        out_dir_files_list = os.listdir(out_path)
        missing_jobs = [i for i in range(num_jobs) if f'{i}.pkl' not in out_dir_files_list]
        are_there_missing_jobs = (len(missing_jobs) > 0)
        if are_there_missing_jobs:
            resent_job_array_ids = _resend_failed_jobs(out_path.parent, missing_jobs, array_name)
            print(f"sent missing jobs: {missing_jobs}")
            sys.stdout.flush()
            job_array_ids += resent_job_array_ids
            num_jobs_to_listen_to += len(missing_jobs)
            is_all_done = False
    print("All output files exist")
    sys.stdout.flush()


def _sum_children_jobs_outputs(num_jobs: int, out_path: Path):
    measure = None
    for j in range(num_jobs):
        with open(os.path.join(out_path, f'{j}.pkl'), 'rb') as f:
            content = pickle.load(f)
            if measure is None:
                measure = content
            else:
                measure += content
    return measure


def cat_children_jobs_outputs(num_jobs: int, out_path: Path, axis: int = 0):
    measure = None
    for j in range(num_jobs):
        with open(os.path.join(out_path, f'{j}.pkl'), 'rb') as f:
            content = pickle.load(f)
            if measure is None:
                measure = content
            else:
                measure = np.concatenate((measure, content), axis=axis)
    return measure


def _check_if_all_done(job_array_ids: list, num_sent_jobs: int) -> bool:
    job_stats_res = subprocess.run(['bjobs', '-A'], stdout=subprocess.PIPE)
    job_stats = job_stats_res.stdout

    # split into different lines, without the header line (and the last empty line which is an artifact of the
    # splitting).
    lines = job_stats.split(b'\n')[1:-1]

    # remove retry lines
    indices_to_remove = []
    for i in range(len(lines)):
        if b'Batch system concurrent query limit exceeded' in lines[i]:
            indices_to_remove.append(i)
    for i in indices_to_remove:
        lines.remove(lines[i])

    # clean spaces
    for i, line in enumerate(lines):
        lines[i] = [c for c in line.split(b' ') if c != b'']

    # Count finished jobs in relevant lines
    num_finished_jobs = 0
    counted_finish_per_line = {}
    array_ids_from_output = [int(lines[i][JOB_ARRAY_ID_IDX]) for i in range(len(lines))]
    for job_arr_id in job_array_ids:
        try:
            relevant_line_idx_in_output = array_ids_from_output.index(job_arr_id)
        except ValueError:
            # There is no line matching this array id in the output of bjobs -A, so it probably hasn't been sent
            # yet.
            print(f"job array {job_arr_id} is not found in bjobs -A output, found ids: {array_ids_from_output}."
                  f"returning False from _check_if_all_done")
            sys.stdout.flush()
            return False
        num_done = int(lines[relevant_line_idx_in_output][DONE_IDX])
        num_exit = int(lines[relevant_line_idx_in_output][EXIT_IDX])
        num_finished_jobs += num_done
        num_finished_jobs += num_exit
        counted_finish_per_line[job_arr_id] = {'done': num_done, 'exit': num_exit}

    if num_finished_jobs < num_sent_jobs:
        return False
    elif num_finished_jobs == num_sent_jobs:
        return True
    else:
        raise ValueError(f"The number of finished jobs {num_finished_jobs} is larger than the total number of "
                         f"sent jobs {num_sent_jobs}, wrong counting!\ncounts per line: {counted_finish_per_line}")


def _resend_failed_jobs(out_path: Path, job_indices: list, array_name: str) -> list:
    num_failed_jobs = len(job_indices)
    job_array_ids = []

    for i in range(num_failed_jobs // LSF_ID_LIST_LEN_LIMIT + 1):
        cur_job_indices_str = ''
        for j_idx_in_list in range(i * LSF_ID_LIST_LEN_LIMIT, min((i + 1) * LSF_ID_LIST_LEN_LIMIT, num_failed_jobs)):
            cur_job_indices_str += f'{job_indices[j_idx_in_list] + 1},'
        cur_job_indices_str = cur_job_indices_str[:-1]
        single_batch_bash_path = os.path.join(out_path, "scripts", "single_batch.sh")
        resend_job_command = f'bsub -J {array_name}[{cur_job_indices_str}]'
        jobs_sending_res = subprocess.run(resend_job_command.split(), stdin=open(single_batch_bash_path, 'r'),
                                          stdout=subprocess.PIPE)
        job_array_ids += _parse_sent_job_array_ids(jobs_sending_res.stdout)

    return job_array_ids


def _parse_sent_job_array_ids(process_stdout) -> list:
    split_array_lines = process_stdout.split(b'\n')[:-1]
    job_array_ids = []
    for line in split_array_lines:
        array_id = int(line.split(b'<')[1].split(b'>')[0])
        job_array_ids.append(array_id)
    return job_array_ids


def generate_binomial_tensor(net_size, node_features_size, num_samples, p=0.5):
    """
    Generate a tensor of size (net_size, net_size, num_samples) where each element is a binomial random variable
    """
    return np.random.binomial(1, p, (net_size, net_size+node_features_size, num_samples)).astype(np.int8)


def sample_from_independent_probabilities_matrix(probability_matrix, sample_size):
    """
    Sample connectivity matrices from a matrix representing the independent probability of an edge between nodes (i, j)
    """
    n_nodes = probability_matrix.shape[0]
    sample = np.zeros((n_nodes, n_nodes, sample_size))

    for i in range(n_nodes):
        for j in range(n_nodes):
            if i == j:
                continue
            sample[i, j, :] = np.random.binomial(1, probability_matrix[i, j], size=sample_size)

    return sample


# TODO: njit?
def split_network_for_bootstrapping(net_size: int, first_part_size: int, splitting_method: str = 'uniform') -> tuple[
    np.ndarray[int], np.ndarray[int]]:
    if splitting_method == 'uniform':
        indices = np.arange(net_size)
        np.random.shuffle(indices)
        first_part_indices = indices[:first_part_size].reshape((first_part_size, 1))
    else:
        raise ValueError(f"splitting method {splitting_method} not supported")

    second_part_indices = np.setdiff1d(np.arange(net_size).astype(int), first_part_indices).reshape(
        (net_size - first_part_size, 1))

    return first_part_indices, second_part_indices
